// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/* ap_fold.C */
/* function for folding */

#ifndef _WIN32
#include "ap_config.h"
#include <unistd.h>
#else
#include <windows.h>
// #define BOINC_APP_GRAPHICS 1 //R: not an option.
#endif

#include <cassert>
#include <cstdio>
#include <cstdlib>

#include "boinc_api.h"
#include "util.h"

#include "astropulse.h"
#include "lcgamm.h"
#include "ap_debug.h"
#include <cmath>

#define ITMAX 1000
#define EPS 1e-6
#define FP_EPS 1.0e-35


namespace Astropulse {

void set_key_nfb(std::vector<int> &key_nfb, float min_freq, int n_bins);

/* General plan: We are given an array of $fold_buf_size power bins.  Each bin contains one float, and
 * corresponds to a sum of powers from
 * $samples_per_bin complex time samples of dechirped data.  In the old fast folding
 * algorithm (henceforth "ffa"), $samples_per_bin was fixed at the number 32.  We now allow it to vary.
 * Hence, $fold_buf_size * $samples_per_bin is a fixed value.
 *
 * We search for repeating signals of various periods T_i.  The period of a signal is defined as
 * the number of power bins over which it repeats.  To do a proper search, we must gradually
 * increment the period in such a way that it looks at every possible period.  This means that we
 * must set T_{i+1} = (1 + 1/$fold_buf_size) T_i every step, so that if two series of pulses,
 * with periods T_{i+1} and T_i, start at the same bin initially, they will end at adjacent bins.
 * The largest period T_0 is set to $fold_buf_size / $freq_0, where $freq_0 = 137, corresponding to
 * a period of 0.1 seconds.  We step by (1 + 1/$fold_buf_size) until we reach T_N = 2T_0,
 * where N is on the order of $fold_buf_size.
 * Then, for each period T_i, we do two things:
 *  1) We use the Fast Folding Algorithm to test periods T_i / 2, T_i / 4, ...
 *  2) We co-add adjacent bins in groups of 2, 4, ...
 * Notes that the most important co-adds (2) are the ones that are around 10% to 30% of the
 * overall period, since this is a typical pulsar duty cycle.
 *
 * Optionally, instead of summing the entire fold buffer, we can sum over one "sub-buffer" at a
 * time, each of which contains $sub_buf_size power bins.  The fold buffer is divided into
 * $fold_buf_size / $sub_buf_size mutually exclusive sub-buffers.  In this case, we increment the period
 * by a factor of (1 + 1/$sub_buf_size) each time, and repeat the procedure for each sub-buffer
 * individually.
 */
inline bool LogRepetitivePulse(float temp_coadd[],int m,int temp_coadd_working_length,int coadd,
				   int n_client_bins,int dm_chunk,int fold_level,bool is_long,
				   int pulse_limit_single, int pulse_limit_rep,double thresh_x_intercept,
				   double thresh_x_slope,double period);
/* Science::ffa
 *
 * arguments:
 * data: array of power bins.  Its length should be state.fold_buf_size, which is
 *       (data bytes) * (samples per byte) / (fold level)
 *    = (8 * 1024 * 1024) * (4) / (state.fold_level_<size>)
 *       where <size> is "large" or "small"
 * num_sub_buffers: we divide the buffer into a number of sub buffers, as discussed above
 * write_thresh_x: if this is set to true, we should be reading a noise file and using it to
 * set the thresholds, which we save in a file.
 */
//FILE* r_debug=NULL;
int Science::ffa(float *data, int num_sub_buffers = 1, bool is_long = true, bool write_thresh_x = false, bool debug_msg = false,
			int pulse_limit_single = 0, int pulse_limit_rep = 0, double ffa_thresh_mult = 1.0) {
#if !((_GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
	Timings<T_ffa> ffa_timing;
#endif
	//ffa_timing.print();
	//Timings<T_dechirp>::print();
	state_t& state = client.ap_shmem->ap_gdata.state;

	/* n_bins = the size of the fold buffer, see above */
	int dm_chunk, n_bins, fold_level;
	double thresh_x_intercept;
	double thresh_x_slope;
	//double thresh_x; //uje: unused variable
	float *temp;
	float *temp_coadd;
#if USE_GPU
	float *gpu_temp;
	float *gpu_temp_coadd;
	float* gpu_data;
#endif
	if(is_long) {
	  dm_chunk = state.dm_chunk_large;
	  n_bins = state.fold_buf_size_long;
	  fold_level = state.fold_level_large;
	  thresh_x_intercept = 4.898856 * ffa_thresh_mult;
	  thresh_x_slope = -0.00000565047 * ffa_thresh_mult;
	  //size_of_temp = size_of_temp_long;
	  temp = temp_long;
	  temp_coadd = temp_coadd_long;
	} else {
	  dm_chunk = state.dm_chunk_small;
	  n_bins = state.fold_buf_size_short;
	  fold_level = state.fold_level_small;
	  thresh_x_intercept = 4.829341 * ffa_thresh_mult;
	  thresh_x_slope = -0.0000536972 * ffa_thresh_mult;
	  //size_of_temp = size_of_temp_short;
	  temp = temp_short;
	  temp_coadd = temp_coadd_short;
	}
#if USE_GPU
	if(is_long) {
	  gpu_temp = gpu_temp_buf_long;
	  gpu_temp_coadd = gpu_temp_coadd_long;
	  gpu_data=gpu_fold_buf_long;
	} else {
	  gpu_temp = gpu_temp_buf_short;
	  gpu_temp_coadd = gpu_temp_coadd_short;
	  gpu_data=gpu_fold_buf_short;
	}
#endif


	//data[n_bins]=0.0f;//R: zeroing guarding element in fold buffer. Now we can remove if(l==n_bins) check in inner loop.
	//JWS: Now allocating enough extra so loops up to 64 floats need no tail handling.
	memset(&data[n_bins], 0, sizeof(float) * 80);

	state.num_sub_buffers = num_sub_buffers;
	int sub_buffer_size = n_bins / state.num_sub_buffers;
	int downup = 0;

#if !NDEBUG
	timer bin_timer;
#endif
//  timer build_table_timer;

	/* For debugging: replaces ap_client_main input with
	 * data that has "ideal" lcgf behavior
	int idum = -1;
	for(int i = 0; i < n_bins; i++) {
	   data[i] = random_combined_power_bin(dm_chunk, idum);
	} */

	/* float freq: the number of periods in a repeating pattern, over the length of the fold buffer.
	 *   (Each fold buffer element representing 2^(fold_level) = 32 samples
	 * int n_per: = (int)freq
	 */
	// using increased precision for period, subfreq, and maxfreq
	double period;
	double subfreq;
	/* Max freq for folding - period of two samples */
	double maxfreq=(double) sub_buffer_size/2.0;
	int per_int, per_int_overlap;
	int j,k,l,m;
	//R: now these arrays in Science itself, properly aligned and padded.
	/*static float *temp_long = NULL;
	static float *temp_short = NULL;
	static float *temp_coadd_long = NULL;
	static float *temp_coadd_short = NULL;
	*/

	int temp_coadd_working_length;
	int max_coadd;
	int n_per_old, n_per = 0;
	int ttidx_pre, ttidx, thresh_tbl[16];
	int n_client_bins_pre_coadd, n_client_bins;
	ap_signal s;
	/* Size is identical to largest used value of per_int in level 1 */
	int size_of_temp = sizeof(float) * (int) ceil((double)sub_buffer_size/state.min_freq);

	/* These are used for one-time allocations, should not depend on passed num_sub_buffers
	 * parameter, but also need be no larger than the maximum possible per_int in level 1.
	 */
//JWS: not used for Lunatics builds
//	if (temp_short == NULL || temp_coadd_short == NULL) size_of_temp_short = sizeof(float) * (int) ceil((double)(state.fold_buf_size_short) / state.min_freq);
//	if (temp_long == NULL || temp_coadd_long == NULL) size_of_temp_long = sizeof(float) * (int) ceil((double)(state.fold_buf_size_long) / state.min_freq);
#if !NDEBUG
	int last_log2 = 0;
#endif
	// using increased precision for dp
	// dp gives the relative period difference per iteration
	double dp;
	dp  = (double)state.num_sub_buffers/(double)n_bins;
	//fprintf(r_debug,"dp=%f\t",dp);
	//fprintf(r_debug,"state.min_freq=%f\n",state.min_freq);
	/* variables that depend on is_long */

//  int *threshold_table; //Jason: No longer needed
//  build_table_timer.start(); //uje: What is this used for in a release version
#if !NDEBUG
	if(!write_thresh_x) {
#endif
#if BOINC_APP_GRAPHICS
	  client.status = "Building ffa threshold table";
#endif
// JWS - big threshold_table replaced by tiny thresh_tbl built at level 1.
//  build_threshold_table(threshold_table, debug_msg, is_long, ffa_thresh_mult);  // Allocates memory for threshold_table
#if !NDEBUG
	} else if(debug_msg) {
	  printf("In ap_fold.cpp: not building threshold table, writing ffa_thresh_mult file instead.\n");
	  fprintf(stderr, "In ap_fold.cpp: not building threshold table, writing ffa_thresh_mult file instead.\n");
	  fflush(stdout);
	}
#endif
//  build_table_timer.stop(); //uje
	//JWS - this still works, code segment was set in main
	if(state.code_segment == build_threshold_table_short) state.code_segment = main_ffa_short;
	else if(state.code_segment == build_threshold_table_long) state.code_segment = main_ffa_long;
	else {
	  fprintf(stderr, "Error: state.code_segment is inconsistent\n");
	  exit(0);  // Error;
	}
#if !NDEBUG
	/* lcgf_file will contain a list of key_nfb's and their corresponding thresholds */
	FILE *thresh_x_file;
	static std::vector<int> key_nfb(0);
	static std::vector<float> max_thresh_x(0);
	static std::vector<int> n_bins_ntc(0);
	std::ostringstream thresh_x_filename;
	int n_ffa_bins;
	// The following vectors will be of the same length:
	static std::vector<int>::iterator iter_key_nfb;  // key_nfb vector
	static std::vector<float>::iterator iter_max_thresh_x; // max_thresh_x vector
	static std::vector<int>::iterator iter_nbins; // num of bins vector
	if(write_thresh_x) {
	  thresh_x_filename << "max_thresh_x_for_dm_chunk_";
	  thresh_x_filename << dm_chunk;
	  thresh_x_file = boinc_fopen((thresh_x_filename.str()).c_str(), "a");
	}

	// Initialize key_nfb, max_thresh_x, n_bins_ntc, only if this is our first run through ffa
	static bool first_run = true;
	if(write_thresh_x && first_run) {
	  set_key_nfb(key_nfb, state.min_freq, n_bins);
	  max_thresh_x.resize(key_nfb.size(), 0);
	  n_bins_ntc.resize(key_nfb.size(), 0);
	  first_run = false;
	}
#endif

	// FILE* writefile = fopen("christest.dat", "a+");
/* R: already allocated in Science calls
	if (temp_long == NULL) {
	  temp_long = (float *)malloc_a(size_of_temp_long);
	  // Should we call an error here if malloc fails?
	}
	if (temp_coadd_long == NULL) {
	  temp_coadd_long = (float *)malloc_a(size_of_temp_long);
	}
	if (temp_short == NULL) {
	  temp_short = (float *)malloc_a(size_of_temp_short);
	}
	if (temp_coadd_short == NULL) {
	  temp_coadd_short = (float *)malloc_a(size_of_temp_short);
	}
*/
	int sb_count = 0;
	long freq_count = 0;
	long subfreq_count = 0;
	long coadd_count = 0;
	long bin_count = 0;
	long pulse_count = 0;
#if (PLENTY_OF_RAM && USE_CONVERSION_OPT)
	static const unsigned int n_per_max=273;
	static const unsigned int large_periods=181721;
	static const unsigned int short_periods=45427;
	typedef unsigned int large_LUT[large_periods][n_per_max];
	typedef unsigned int short_LUT[short_periods][n_per_max];
	int large_offsets_file,short_offsets_file;
	static large_LUT* large_offsets_LUT=NULL;
	static short_LUT* short_offsets_LUT=NULL;
	if(is_long){
	  if(!large_offsets_LUT){
		large_offsets_LUT=(large_LUT*)malloc_a(large_periods*n_per_max*sizeof(unsigned int));
		if(!large_offsets_LUT){
		  fprintf(stderr,"Error: Can't allocate offsets lookup table\n");
		  exit(-1);
		}
		if((large_offsets_file=open("c:\\ap_large_LUT.sah",_O_RDONLY|_O_SEQUENTIAL|_O_BINARY ))==-1){
		  fprintf(stderr,"Generating new lookup table\n");
		  if((large_offsets_file=open("c:\\ap_large_LUT.sah",_O_WRONLY | _O_CREAT|_O_SEQUENTIAL|_O_BINARY))==-1){
			fprintf(stderr,"Error: Can't create lookup table file\n");
			exit(-1);
		  }
		  float freq=(float)state.min_freq;//should be inside loop, but unallowed syntaxis...
		  for(unsigned freq_i=0;freq< 2*(float)state.min_freq; freq *= (1 + dp),freq_i++){
			register const float period= (float)sub_buffer_size/freq;
			for(unsigned per_i=0;per_i<(unsigned)freq;per_i++) (*large_offsets_LUT)[freq_i][per_i]= (unsigned)(period*(float)per_i + 0.5);
		  }
		  write(large_offsets_file,large_offsets_LUT,large_periods*n_per_max*sizeof(unsigned int));
		}else{//file exists
		  read(large_offsets_file,large_offsets_LUT,large_periods*n_per_max*sizeof(unsigned int));
		}
		close(large_offsets_file);
	  }
	} else {
	  if(!short_offsets_LUT){
		short_offsets_LUT=(short_LUT*)malloc_a(short_periods*n_per_max*sizeof(unsigned int));
		if(!short_offsets_LUT){
		  fprintf(stderr,"Error: Can't allocate offsets lookup table\n");
		  exit(-1);
		}
		if((short_offsets_file=open("c:\\ap_short_LUT.sah",_O_RDONLY|_O_SEQUENTIAL|_O_BINARY ))==-1){
		  fprintf(stderr,"Generating new lookup table\n");
		  if((short_offsets_file=open("c:\\ap_short_LUT.sah",_O_WRONLY | _O_CREAT|_O_SEQUENTIAL|_O_BINARY))==-1){
			fprintf(stderr,"Error: Can't create lookup table file\n");
			exit(-1);
		  }
		  float freq=(float)state.min_freq;
		  for(unsigned freq_i=0;freq< 2*(float)state.min_freq; freq *= (1 + dp),freq_i++){
			register const float period= (float)sub_buffer_size/freq;
			for(unsigned per_i=0;per_i<(unsigned)freq;per_i++) (*short_offsets_LUT)[freq_i][per_i]= (unsigned)(period*(float)per_i + 0.5);
//			write(short_offsets_file,&((*short_offsets_LUT)[freq_i][per_i]),sizeof(unsigned int));
//			fprintf(r_debug,"LUT[%u][%u]=%u\n",freq_i,per_i,(*short_offsets_LUT)[freq_i][per_i]);
//			fprintf(r_debug,"\n");
		  }
		  write(short_offsets_file,short_offsets_LUT,short_periods*n_per_max*sizeof(unsigned));
		}else{//file exists
		//read(short_offsets_file,short_offsets_LUT,short_periods*n_per_max*sizeof(int));
//		  float freq=(float)state.min_freq;
//		  for(int freq_i=0;freq< 2*(float)state.min_freq; freq *= (1 + dp),freq_i++){
//			for(unsigned per_i=0;per_i<(unsigned)freq;per_i++){
//			  unsigned tmp=0;
//			  _read(short_offsets_file,&tmp,sizeof(unsigned));
//			  fprintf(r_debug,"LUT_read[%u][%u]=%u\n",freq_i,per_i,tmp);
			  read(short_offsets_file,short_offsets_LUT,short_periods*n_per_max*sizeof(unsigned));
//			  (*short_offsets_LUT)[freq_i][per_i]=tmp;
//			}
//		  }
		}
		close(short_offsets_file);
	  }
	}

#endif
	/* Level 0: Loop over all sub-buffers */
#if 0
	fprintf(stderr,"FFA input data:\n");
#endif
	for(; state.sub_buffer < state.num_sub_buffers; state.sub_buffer++) {
	  sb_count++;
#if BOINC_APP_GRAPHICS
	  client.status= "Computing fast fold.";
#endif
#if 1
	unsigned int tID=0;
#endif
	  if(state.freq == 0) state.freq = (double)state.min_freq;
	  /* Level 1: Loop over all periods, from min to max */
	  for(; state.freq < 2*(double)state.min_freq; state.freq *= (1 + dp)) {
		double cur_cpu_time;
#if BOINC_APP_GRAPHICS
		client.status = "Fast folding algorithm";
#endif
#ifndef NDEBUG
		/* Report to stdout if we have reached a new "milestone" frequency. */
		if(int_log2(freq_count) > last_log2) {
		  if(debug_msg) {
			printf("In ap_fold.cpp: freq = %f\n", state.freq);
			fprintf(stderr, "In ap_fold.cpp: freq = %f\n", state.freq);
			fflush(stdout);
		  }
		  last_log2 = int_log2(freq_count);
		}
#endif
		/* Set the period we will consider */
		period = (double)(sub_buffer_size/state.freq);
		//if(is_long)fprintf(r_debug,"%f\t",period);
		/* $per_int is the number of bins we will need in the temp array; one for
		 * each bin in a period, and rounded up so we can catch the last fraction of
		 * a bin.
		 */
		per_int = (int)ceil(period);
#if 1
#if !((_GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
	//	if(is_long) Counters<T_per_int_long,long>::update(per_int);
	//	else Counters<T_per_int_short,long>::update(per_int);
#endif
#endif
		/* $n_per is the number of discrete blocks of length $period in one sub-buffer.
		 * We will build the @temp array by looping through all $n_per blocks,
		 * and adding each of them to @temp.
		 */
		n_per_old = n_per;
		n_per = (int)state.freq;
		/*if(is_long)fprintf(r_debug,"n_per=%d\n",n_per);*/
		/* Number of bins of original @power array that we are adding together to
		 * make each bin of @temp.
		 */
		n_client_bins_pre_coadd = dm_chunk*n_per;
		// build small table of thresholds as needed
		if (n_per != n_per_old) {
		  max_coadd = int_log2(per_int);
		  int nfb;  //Jason: Had to pull these out of for loop init section :C, Blame MS
		  int ncb;
		  for ( ttidx_pre = 0,nfb = n_per,ncb = n_client_bins_pre_coadd; ttidx_pre <= max_coadd; ttidx_pre++, nfb<<=1, ncb<<=1 ) {
			double thresh_x = thresh_x_intercept + thresh_x_slope * nfb;
			float threshtmp = ncb + thresh_x * sqrt(2.0 * (float) ncb);
			thresh_tbl[ttidx_pre] = (int)threshtmp;
//			fprintf(stderr,"Creating: threshtbl[%d] = %d\n",ttidx_pre,thresh_tbl[ttidx_pre]);
		  }

		  /* boinc API boinc_fraction_done tells BOINC how much of the current workunit has
		   * been completed. This is used to inform the core client GUI of the % complete.
		   * JWS: now doing so only 137 times per FFA rather than up to 181704 times.
		   */
		  state.compute_fraction_done(-1);
#if  !(USE_CONVERSION_OPT && PLENTY_OF_RAM)
		  client.checkpoint();
#endif
		  boinc_fraction_done(state.frac_done);
//		  boinc_ops_cumulative(state.frac_done*FLOPS_PER_DM*(state.dm_hi-state.dm_low)*log((double)state.fft_len)/10.39720770839917918237915728241,0);//log(32768.0)
/*uje:  boinc_ops_cumulative() was deprecated some time ago and has been removed with BOINC changeset 25700!
  boinc_ops_cumulative(state.frac_done*fpopsMult, 0);
*/
		  boinc_wu_cpu_time(cur_cpu_time);
		  client.ap_shmem->ap_gdata.cpu_time = cur_cpu_time;
		}
		ttidx_pre = 0; //Jason: moved this out of if block above
		/* Set all elements of the temp buffer to zero */
#if USE_GPU
		cudaMemset( gpu_temp,0, size_of_temp );
#else
		memset(temp, 0, size_of_temp+128&~63);
		memset(temp_coadd, 0, size_of_temp+128&~63);
		memset(temp_o2, 0, size_of_temp+128&~63);
		memset(temp_o3, 0, size_of_temp+128&~63);
#endif
		/* Fill the temp buffer.  How this works: define
		 * A float block: a span of consecutive bins with a possibly non-integer length equal
		 *   to $period.  It can start and end at fractions of bins.  I visualize the bins
		 *   as having a left edge and a right edge, and a length of 1 unit.  E.g. bin 0
		 *   has a left edge at 0, and a right edge at 1.
		 * An int block: an discrete set of bins of length $per_int (which is $period rounded up)
		 *   in the fold buffer.  By discrete, I mean that it starts at the left edge of its
		 *   leftmost bin, and ends at the right edge of its rightmost bin.
		 * Then this code effectively divides the subbuffer into nonoverlapping, consecutive
		 * float blocks and (at the same time) possibly overlapping int blocks.
		 * Any final (rightmost) bins that are too short to form a block of length $period
		 * are ignored when dividing into float blocks.
		 * The code then adds the int blocks together elementwise to make the @temp array.
		 * Each int block
		 * corresponds to one float block, and it starts in the bin such that the left edge of
		 * the int block is as close as possible to the left edge of the float block.
		 *
		 */
#if (USE_CONVERSION_OPT && PLENTY_OF_RAM) //uje: unused by gcc
//R: state.sub_buffer always==0 cause  num_sub_buffers == 1 in all FFA calls. So I removed non needed addition to l
		if(is_long){
//R:rearraging loops to achieve better cache locality
		  for (k=0; k<n_per; k++) {
			register int ll=(*large_offsets_LUT)[freq_count][k];
  #if __SSE__
			if(!(per_int%4 || ll%4)){//R: we can use aligned memory access
			  for (j=0; j<per_int; j+=4) _mm_store_ps(&temp[j],_mm_add_ps(_mm_load_ps(&temp[j]),_mm_load_ps(&data[ll+j])));
			}else{
  #endif
			  for (j=0; j<per_int; j++) {
				l = ll + j;
				//if(l != n_bins)
				temp[j] += data[l];
				//else fprintf(r_debug,"Potential error l==n_bins\n values are n_bins=%d, j=%d, k=%d, n_per=%d, per_int=%d\n",n_bins,j,k,n_per,per_int);
			  }
  #if __SSE__
			}//checking aligment
  #endif
		  }
		}else{//short
		  for (k=0; k<n_per; k++) {
			register int ll=(*short_offsets_LUT)[freq_count][k];
  #if __SSE__
			if(!(per_int%4 || ll%4)){//R: we can use aligned memory access
			  for (j=0; j<per_int; j+=4) {
				_mm_store_ps(&temp[j],_mm_add_ps(_mm_load_ps(&temp[j]),_mm_load_ps(&data[ll+j])));
			  }
			}else{
  #endif
			  for (j=0; j<per_int; j++) {
				l = ll + j;
				//if(l != n_bins)
				temp[j] += data[l];
				//else fprintf(r_debug,"Potential error l==n_bins\n values are n_bins=%d, j=%d, k=%d, n_per=%d, per_int=%d\n",n_bins,j,k,n_per,per_int);
			  }
  #if __SSE__
			}//checking aligment
  #endif
		  }//k-loop
		}//end short

#else
  #if (USE_CONVERSION_OPT && !PLENTY_OF_RAM && !USE_LRINT) //uje: used by gcc
		int* offsets=new int[n_per];
		//register const int buf_size=state.sub_buffer * sub_buffer_size; //R: Always zero!
		int i; //uje: compiler says new rules
		for(i=0;i<n_per;i++)
	#if 0 //__SSE2__
		  {
		  float tmp=period*(float)i + 0.5;
		  offsets[i]=_mm_cvt_ss2si(_mm_load_ss(&f));
		}
	#else
		  {
		  offsets[i]=(int)(period*(float)i + 0.5);//+ buf_size;
		}
	#endif
  #endif

  #if USE_GPU
		for (k=0; k<n_per; k++) {
		  register int ll=offsets[k];
		  GPU_fetch_array(gpu_temp,&gpu_data[ll],per_int);
		}
  #elif (__SSE__ && USE_CONVERSION_OPT && !PLENTY_OF_RAM) //uje: used by gcc
	#if USE_LRINT
		//JWS: MOD_20 imported from unit testing
		__m128 rg1, rg2, rg3, rg4;
		float * sumdst[4] = {temp, temp_coadd, temp_o2, temp_o3};
		unsigned ll, ji;
		int kk, kkchg;
		if (downup++ & 1) {
		  kk = 0;
		  kkchg = 1;
		} else {
		  kk = n_per - 1;
		  kkchg = -1;
		}
		for ( ;(unsigned)kk < (unsigned)n_per; kk += kkchg) {
		  ll = (unsigned)lrint(period*kk);
		  float * tP = sumdst[ll&3];
		  ll &= ~3;
		  for(ji = 0 ; ji < per_int+3; ji += 8) {
			rg1 = _mm_load_ps(&data[ll+ji]);
			rg2 = _mm_load_ps(&data[ll+ji+4]);
			rg3 = _mm_load_ps(&tP[ji]);
			rg4 = _mm_load_ps(&tP[ji+4]);
			rg3 = _mm_add_ps(rg3,rg1);
			rg4 = _mm_add_ps(rg4,rg2);
			_mm_store_ps(&tP[ji],rg3);
			_mm_store_ps(&tP[ji+4],rg4);
		  }
		} //kk
		for(ji = 0 ; ji < per_int; ji += 4) {
		  rg4 = _mm_loadu_ps(&temp_o3[ji+3]);
		  rg3 = _mm_loadu_ps(&temp_o2[ji+2]);
		  rg2 = _mm_loadu_ps(&temp_coadd[ji+1]);
		  rg1 = _mm_load_ps(&temp[ji]);
		  rg2 = _mm_add_ps(rg2,rg4);
		  rg1 = _mm_add_ps(rg1,rg3);
		  _mm_store_ps(&temp[ji],_mm_add_ps(rg2,rg1));
		} //ji

	#elif !(_GNU_SOURCE && (__WORDSIZE == 32))
		for (k=0; k<n_per; k++) {
		  register int ll=offsets[k];
		  if (ll&3) {//R:store and temp loads aligned only
			for(j=0;j<per_int;j+=16) {
			  _mm_store_ps(&temp[j],_mm_add_ps(_mm_load_ps(&temp[j]),_mm_loadu_ps(&data[ll+j])));
			  _mm_store_ps(&temp[j+4],_mm_add_ps(_mm_load_ps(&temp[j+4]),_mm_loadu_ps(&data[ll+j+4])));
			  _mm_store_ps(&temp[j+8],_mm_add_ps(_mm_load_ps(&temp[j+8]),_mm_loadu_ps(&data[ll+j+8])));
			  _mm_store_ps(&temp[j+12],_mm_add_ps(_mm_load_ps(&temp[j+12]),_mm_loadu_ps(&data[ll+j+12])));
			}
		  } else {//R: we can use aligned memory access for all
			for(j=0;j<per_int;j+=16) {
			  _mm_store_ps(&temp[j],_mm_add_ps(_mm_load_ps(&temp[j]),_mm_load_ps(&data[ll+j])));
			  _mm_store_ps(&temp[j+4],_mm_add_ps(_mm_load_ps(&temp[j+4]),_mm_load_ps(&data[ll+j+4])));
			  _mm_store_ps(&temp[j+8],_mm_add_ps(_mm_load_ps(&temp[j+8]),_mm_load_ps(&data[ll+j+8])));
			  _mm_store_ps(&temp[j+12],_mm_add_ps(_mm_load_ps(&temp[j+12]),_mm_load_ps(&data[ll+j+12])));
			}
		  }
		}//k
	#elif (_GNU_SOURCE && (__WORDSIZE == 32)) //uje: unaligned loads on 32bit linux only
		for (k=0; k<n_per; k++) {
		register int ll=offsets[k];
		for(int j=0;j<per_int;j+=4) _mm_store_ps(&temp[j],_mm_add_ps(_mm_loadu_ps(&temp[j]),_mm_loadu_ps(&data[ll+j])));
		}//k
	#endif

  #else // !__SSE__,
		for(k=0; k<n_per; k++) {
		  for (j=0; j<per_int; j++) {
	#if (USE_CONVERSION_OPT && !PLENTY_OF_RAM) //uje: used by gcc
			register int ll=offsets[k];
			l = ll + j;
	#else
			l = (int)(period*(float)k + 0.5 + (float)j) + state.sub_buffer * sub_buffer_size;
	#endif //USE_CONVERSION...
			/* Note: there's a danger that l == n_bins and we'll go over the edge of the
			 * fold buffer.  This would happen if
			 * a) period == an integer + frac, for some frac < 0.5
			 * b) period almost exactly divide n_bins, such that period * n_per > n_bins - 0.5 + frac
			 * By (a), we have per_int = period + 1 - frac, so when j and k are maximum,
			 * l = (int)(x), where
			 *   x = period * (n_per - 1) + (per_int - 1) + 0.5
			 *   x = (period * n_per - frac + 0.5)
			 *   x > n_bins
			 * l = (int)(x) >= n_bins
			 * l can't actually by > n_bins, so it must equal n_bins if this happens.  Hence
			 * the "if" clause in the following statement:
			 */
			//if(l != n_bins)
			temp[j] += data[l];
		  }//j
		}//k
  #endif
  #if (USE_CONVERSION_OPT && !PLENTY_OF_RAM && !USE_LRINT) //uje: used by gcc
		delete[] offsets;
  #endif
#endif // (USE_CONVERSION_OPT && PLENTY_OF_RAM)
		freq_count++;
		/* Level 2: Loop over all subfrequencies freq, freq*2, freq*4 ... */
		for(subfreq = state.freq; subfreq < maxfreq; subfreq *= 2) {
		  subfreq_count++;
		  /* Note that per_int is not adjusted here, but at the end of the for loop. */
		  period =(double)sub_buffer_size/subfreq;
		  /* We can coadd up to 2^max_coadd bins, which must be at most half the period. */
		  max_coadd = int_log2(per_int) - 1;
		  /* Create a second temporary buffer to hold the co-added bins.  */
#if USE_GPU
		  cudaMemcpy(gpu_temp_coadd,gpu_temp,sizeof(float) * per_int,cudaMemcpyDeviceToDevice);
#else
		  memcpy(temp_coadd, temp, sizeof(float) * per_int);
#endif
		  /* The size of the coadd buffer starts out equal to a period, then shrinks by factors of 2. */
		  temp_coadd_working_length = per_int;
		  /* n_client_bins_pre_coadd is the number of client bins summed at coadd 0, and will double
		   *   each time we cycle through the level 2 loop.  (Iterate through doubled frequencies)
		   * n_client_bins is the number of client bins summed at a particular coadd, and will double
		   *   each time we cycle through the level 3 loop, in addition to the level 2 loop.
		   * The doubling happens at the end of the loop.
		   */
		  n_client_bins = n_client_bins_pre_coadd;
		  ttidx = ttidx_pre;
		  /* Level 3: Loop over coadds */
		  for(int coadd = 0; coadd <= max_coadd; coadd++ ) {
			coadd_count++;
#if !NDEBUG
		  if(write_thresh_x) {
			n_ffa_bins = n_client_bins / dm_chunk;
			/* Find the first key_nfb that is larger than n_ffa_bins, and
			 * prepare iter_max_thresh_x to point to the corresponding max_thresh_x
			 */
			for(iter_key_nfb = key_nfb.begin(), iter_max_thresh_x = max_thresh_x.begin(), iter_nbins = n_bins_ntc.begin()
			  ;iter_key_nfb != key_nfb.end()
			  ; iter_key_nfb++, iter_max_thresh_x++, iter_nbins++) {
			  assert(iter_max_thresh_x != max_thresh_x.end() && iter_nbins != n_bins_ntc.end());
			  if(*iter_key_nfb > n_ffa_bins) break;
			}
			int q = *iter_key_nfb;  // Pause point for gdb
			/* Count the number of bins per key_nfb */
			*iter_nbins += temp_coadd_working_length;
		  }
#endif
		  /* Level 4: Loop over bins.  We'll do the coadd at the end of the loop. */
#if !NDEBUG
		  bin_timer.start();
#endif
//Jason: Playing
//		  register float threshold_orig = threshold_table[n_client_bins / dm_chunk];
//		  fprintf(stderr,"ttidx = %d\n",ttidx); //Jason Debugging
		  float threshold = thresh_tbl[ttidx]; //JWS - minimal table
//		  fprintf(stderr,"Using thresh_tbl[%u] = %f\n",ttidx,threshold);
{
#if !((_GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
//Timings<T_FFA_compare> count;
#endif
#if USE_GPU
		  bool need_CPU_compare=false;
		  GPU_compare_with_threshold(gpu_temp_coadd,temp_coadd_working_length,threshold,need_CPU_compare);
		  if(need_CPU_compare){
			cudaMemcpy(temp_coadd,gpu_temp_coadd,sizeof(float)*temp_coadd_working_length,cudaMemcpyDeviceToHost);
#endif
		  for(m = 0; m < temp_coadd_working_length; m++) {
			bin_count++;
			/* If a bin goes over threshhold, record it.
			 * We determine if a bin is over threshhold by deciding whether its
			 * probability of existence (in random noise) is < 1/(number of bins we examine per workunit)
			 * This ensures that on average we will detect 1 false positive.
			 * There are N freq's, where (1 + 1/sub_buffer_size)^N = 2.
			 * That is, e^(N/sub_buffer_size) = 2, so N = sub_buffer_size * ln 2.
			 * Each freq has a period equal to sub_buffer_size / freq bins, for a
			 * total of about sub_buffer_size^2 / state.min_freq terms.
			 * And sub_buffer_size = num_samples / (dm_chunk * state.num_sub_buffers), so that makes
			 * num_samples^2 / (dm_chunk^2 * state.num_sub_buffers^2 * state.min_freq)
			 * Second, for each freq, there are a number of subfreqs, each of which has
			 * half as many bins as the previous.  This doubles the number of bins.
			 * Third, for each subfreqs, there are a number of coadds, each of which
			 * has half as many bins as the previous.  This doubles the number of
			 * bins again.
			 * Fourth, the code runs state.num_sub_buffers times, so we multiply by that.
			 * Finally, there are 15000 / dm_chunk dm chunks, and we run
			 * the ffa once per dm chunk.
			 * Therefore, the number of bins we examine per workunit is
			 * (num_samples^2 * 4 * 15000) / (state.num_sub_buffers * state.min_freq * dm_chunk^3).
			 * Currently, num_samples = 32 * 10^6, state.num_sub_buffers = 1, state.min_freq = 137,
			 * and dm_chunk = 128.
			 * That's 2 * 10^11 bins per workunit, which is e^26.18
			 * This suggests that the lcgf should have a threshold of -26.  But evidently this isn't enough ... ?
			 * Note: for dm_chunk_small, change sub_buffer_size to fold_buf_bytes_short * 4 / dm_chunk
			 * (fold_buf_bytes_short^2 * 16 * 4 * 15000) / (state.num_sub_buffers * state.min_freq
			 *                * dm_chunk^3)    ... I get e^(-25.5)
			 *
			 * First, total power in these bins must exceed the number of client bins.
			 */
			//if(temp_coadd[m] > n_client_bins + 1) { R:Unneeded check.
			/* If we're writing the thresholds file, then
			 * check if we have a new max thresh_x for this key_nfb.
			 */
#if !NDEBUG
			if (write_thresh_x) {
			  float cur_thresh_x = (temp_coadd[m] - n_client_bins) / sqrt(2 * (double) n_client_bins);
			  // lcgf((double) n_client_bins, temp_coadd[m]);
			  if(cur_thresh_x > *iter_max_thresh_x) *iter_max_thresh_x = cur_thresh_x;
			} else
#endif
			if (temp_coadd[m] >= threshold) {
			  // Temporary split between if statements removed
			  pulse_count++;

			  time_series_b_t time_series;
			  time_series.build(temp_coadd, m, temp_coadd_working_length, (double) n_client_bins);

			  /* I don't understand why state.dm_now / dm_chunk is here. */
			  state.pulse_jd = state.jd0 + (state.data_chunk_now + state.dm_now/dm_chunk) /(state.datasize)*(state.jd1-state.jd0);
			  s.time=state.pulse_jd;
			  // s.peak_bin = state.data_chunk_now*4 + m;
			  s.peak_bin = m * dm_chunk * (1 << coadd);
			  s.peak_power = temp_coadd[m];
			  s.scale = fold_level;
			  if(is_long) s.dm = state.dm_sign*(state.dm_chunk_large_now);
			  else s.dm = state.dm_sign*(state.dm_chunk_large_now+state.dm_chunk_small_now);
			  s.period = period;
			  s.ffa_scale = coadd;
			  // Machinery for writing s.num_std_devs.  fudge_factor is about 1.
			  float fudge_factor;
			  fudge_factor = 1 - thresh_x_slope * (n_client_bins>>fold_level) / thresh_x_intercept;
			  s.num_std_devs = (s.peak_power - n_client_bins)/(fudge_factor * sqrt((double)n_client_bins));
			  s.time_series=time_series.data;
			  if(add_signal(client.signal_vector, s)) state.result_count_rep++;
			   // client.outfile.output_result(s);
			  if(pulse_limit_rep && state.result_count_rep >= pulse_limit_rep) {
				// We've already found enough repeating pulses.
				if(pulse_limit_single && state.result_count_single >= pulse_limit_single) {
				// Quit early
				  printf("Found %d single pulses and %d repeating pulses, exiting.\n", state.result_count_single, state.result_count_rep);
				  fprintf(stderr, "Found %d single pulses and %d repeating pulses, exiting.\n", state.result_count_single, state.result_count_rep);
				  fflush(stdout);
				  client.outfile.write_main(client.signal_vector);
				  client.outfile.finish_results();
				  client.finish();
				  main_timer.stop();
				  boinc_finish(0);
				  exit(0);
				} else {
				  //(Must be that we haven't found enough single pulses.)
				  // Don't waste time searching for more repetitive pulses.
				  skipffa_short = true;                     // bypass shorts
				  skipffa_long = true;                      // bypass longs
				  state.fold_level_small = 99;              // inhibit short save
				  state.fold_level_large = 99;              // inhibit long save
				  return(0);
				}
			  }
			}// End if (bin goes over threshhold)
			// } R: unneeded check elimination
		  } // End loop over bins
#if USE_GPU
			}//R:need to check array for pulses only if GPU found pulse already
#endif
}//R:compare block end
#if !NDEBUG
			bin_timer.stop();
#endif
			/* Coadd adjacent bins, and throw away the odd bin at the end if any. */

			/* Now that our coadd has increased, there are less bins in our temp_coadd array ... */
			temp_coadd_working_length >>= 1;
{
#if !((_GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
//Timings<T_FFA_coadd> count;
#endif
#if USE_GPU
			GPU_coadd_in_place(gpu_temp_coadd, per_int);
#else
			Science::coadd_in_place(temp_coadd, per_int);
#endif
}//R: coadd block end

			/* ... and there are more samples represented by each bin */
			n_client_bins <<= 1;
			ttidx++;
		  } // End loop over coadds

		  /* Cut the temp array's size in half, adding the second half to the first half
		   *
		   * Note per_int and per_int_overlap differ iff the old per_int is odd.
		   *
		   * Odd block size issues:
		   * There's some question about what to do when the array has an odd size.
		   * I believe the most sensible thing to do is to round up, and allow the
		   * two halves to overlap by one bin.  This is in keeping with our convention that
		   * $per_int is $period rounded up.  If $per_int is odd, then $period is slightly
		   * greater than an even integer, so $period / 2 is also slightly greater than
		   * an even integer.  Therefore the new $per_int should be odd as well.
		   * However, no system is entirely satisfactory.  Under certain conditions, it is
		   * possible for float blocks to differ from their respective int blocks by as
		   * much as 1.5 bins.  This may interfere with detection of very narrow pulses.
		   * (To see this effect, try starting with block of size 9, whose float block is
		   * of length 9 - epsilon and starts at 0.5 - epsilon after the int block starts.
		   * Cut the int block in halves twice and look at the last of the four subblocks.)
		   */
		  per_int_overlap = per_int >> 1;
		  per_int = (per_int + 1)  >> 1;
{
#if !((_GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
//Timings<T_FFA_stride_add> count;
#endif
#if USE_GPU
		  GPU_coadd_with_stride(gpu_temp,per_int,per_int_overlap);
#elif __SSE__
		  __m128 vs1, vs2, vd1, vd2;
		  for (j=0; j<per_int; j+=8) {
			vs1 = _mm_loadu_ps(&temp[j+per_int_overlap]);
			vs2 = _mm_loadu_ps(&temp[j+per_int_overlap+4]);
			vd1 = _mm_load_ps(&temp[j]);
			vd2 = _mm_load_ps(&temp[j+4]);
			vd1 = _mm_add_ps(vd1, vs1);
			vd2 = _mm_add_ps(vd2, vs2);
			_mm_store_ps(&temp[j], vd1);
			_mm_store_ps(&temp[j+4], vd2);
		  }
#else
		  for (j=0; j<per_int; j++)  temp[j] = temp[j] + temp[j+per_int_overlap];
#endif
}//R: stride coadd block end
		  n_client_bins_pre_coadd = n_client_bins_pre_coadd << 1;
		  ttidx_pre++;
		} // End loop over subfrequencies
	  } // End loop over periods
	  //fprintf(r_debug,"freq_count=%d\n",freq_count);
	  state.freq = 0;  // Codes for min freq
	} // End loop over all subbuffers
	state.sub_buffer = 0;

//  if(!write_thresh_x) {  // WARNING: if we're writing thresh_x, this is never allocated to begin with.
//    if(threshold_table) {delete [] threshold_table; threshold_table = NULL;}
//  }
#if !NDEBUG
	if(write_thresh_x) {
	  for(iter_key_nfb = key_nfb.begin(), iter_max_thresh_x = max_thresh_x.begin(), iter_nbins = n_bins_ntc.begin()
		  ;iter_key_nfb != key_nfb.end()
		  ; iter_key_nfb++, iter_max_thresh_x++, iter_nbins++) {
		assert(iter_max_thresh_x != max_thresh_x.end() && iter_nbins != n_bins_ntc.end());
		fprintf(thresh_x_file, "%d %f %d\n", *iter_key_nfb, *iter_max_thresh_x, *iter_nbins);
	  }
	  fclose(thresh_x_file);
	}

	if(debug_msg) {
	  //printf("Time spent in building threshold table: %d seconds, %d ticks\n", build_table_timer.get_seconds(), build_table_timer.get_ticks()); //uje: unused
	  //fprintf(stderr, "Time spent in building threshold table: %d seconds, %d ticks\n", build_table_timer.get_seconds(), build_table_timer.get_ticks()); //uje: unused
	  printf("Time spent in bins: %d seconds, %d ticks\n", bin_timer.get_seconds(), bin_timer.get_ticks());
	  fprintf(stderr, "Time spent in bins: %d seconds, %d ticks\n", bin_timer.get_seconds(), bin_timer.get_ticks());
	  printf("sb_count: %d\nfreq_count: %ld\nsubfreq_count: %ld\ncoadd_count: %ld\nbin_count: %ld\npulse_count: %ld\n",
		  sb_count, freq_count, subfreq_count, coadd_count, bin_count, pulse_count);
	  fprintf(stderr, "sb_count: %d\nfreq_count: %ld\nsubfreq_count: %ld\ncoadd_count: %ld\nbin_count: %ld\npulse_count: %ld\n",
		  sb_count, freq_count, subfreq_count, coadd_count, bin_count, pulse_count);
	  fflush(stdout);
	}
#endif
	return(0);
} //Science::ffa(...)

#if 0  //JWS - table replaced by calculation
/* build_threshold_table
 *
 * Create an array, and store it in @table.
 * The index of the array represents an nfb ("number of ffa bins"), and the value represents
 * the threshold for that nfb.  If we add $nfb ffa bins together, and the total power has
 * power > table[nfb], we have detected a pulse.
 */
void Science::build_threshold_table(int*& table, bool debug_msg, bool is_long, double ffa_thresh_mult) {
	double thresh_x_intercept;
	double thresh_x_slope;
	double thresh_x;

	// Numbers come from lcgf_test.cpp, using e^-25.49 for short and e^-26.18
	// for long fold buffers
	// Note this (below) doesn't mean the x intercept, it means the y intercept for the
	// function thresh_x(nfb).  Bad notation.
	if(is_long) {
	  // rewrite this one
	  // artificially high one: thresh_x_intercept = 10;
	  thresh_x_intercept = 4.898856 * ffa_thresh_mult;
	  // old: thresh_x_slope = -0.0000131539;
	  thresh_x_slope = -0.00000565047 * ffa_thresh_mult;
	} else {
	  // artificially high one: thresh_x_intercept = 10;
	  thresh_x_intercept = 4.829341 * ffa_thresh_mult;
	  thresh_x_slope = -0.0000536972 * ffa_thresh_mult;
	}

	state_t& state = client.ap_shmem->ap_gdata.state;
	int min_nfb = state.min_freq; // We add at least min_freq ffa bins ...
	int max_nfb;
	if(is_long)
	  max_nfb = state.fold_buf_size_long; // And at most all of the bins.
	else
	  max_nfb = state.fold_buf_size_short;
	float threshold;
#if !NDEBUG
	if (debug_msg) {
	  printf("In ap_fold.cpp: at Science::build_threshold_table()\n");
	  fprintf(stderr, "In ap_fold.cpp: at Science::build_threshold_table()\n");
	  fflush(stdout);
	}
#endif
	table = new int[max_nfb+1];

	for(state.nfb = min_nfb; state.nfb <= max_nfb; state.nfb++) {
	  int ncb;
	  if(is_long) {ncb = state.nfb * state.dm_chunk_large;}  // Are we in long or short ffa?
			 else {ncb = state.nfb * state.dm_chunk_small;}
	  state.compute_fraction_done(-2);
	  thresh_x = thresh_x_intercept + thresh_x_slope * state.nfb;
	  threshold = ncb + thresh_x * sqrt(2.0 * (double) ncb);
	  table[state.nfb] = threshold;
	}
}
#endif

/* Search for the smallest power_sum between min and max inclusive
 * such that lcgf(ncb, power_sum) <= threshold.  If no value satisfies
 * the inequality, return -1, a signal that we should search higher.
 *
 * Since we are no longer using lcgf, this function is no longer in use
 */
/*
int Science::lcgf_binary_search(int min, int max, int ncb, float thresh_lcgf) {
	if (min == max) { // base case
	  if (lcgf(ncb, min) > thresh_lcgf) return -1; // Search higher!
	  else return min; // Found it!
	}
	int cut = (min + max) / 2;
	if (lcgf(ncb, cut) <= thresh_lcgf) return lcgf_binary_search(min, cut, ncb, thresh_lcgf);
	else return lcgf_binary_search(cut + 1, max, ncb, thresh_lcgf);
}
*/
/* set_key_nfb
 *
 * Arguments:
 * key_nfb: output array  (nfb = number of ffa bins)
 * min_freq: minimum frequency considered by the ffa
 * n_bins: number of ffa bins processed by the ffa (each ffa bin corresponds to dm_chunk of the client's bins)
 *
 * Algorithm:
 * The purpose of this function is to set an array of key_nfb's.  An nfb is a number of ffa bins
 * that we add together to make one power sum.  Between any two consecutive key_nfbs should be roughly
 * the same number of power sums.  (That way, when we find the min lcgfs for each key_nfb, they will be
 * roughly the same.)
 * For instance, if the array of key_nfbs goes: 137, 139, 142 ...  then we would expect there to be
 * the same number of power sums between 137 to 139 as between 139 to 142.
 * Hence, to perform this algorithm we must be able to count how many power sums will be processed
 * for each ncb.  (Number of client bins.)
 *
 * The first point is that the only allowed n_ffa_bins are of the form base_freq * 2^n, where
 * min_freq <= base_freq < 2 * min_freq.  However, we treat n_ffa_bins as a continuous variable.
 * For each n_ffa_bins, we estimate the number of power sums between it and n_ffa_bins + 1.
 *
 * This is the product of:
 *
 *  1) Assume we're at coadd 0, so n_ffa_bins = subfreq.  Then find # of subfreqs in the interval:
 *  subfreqs = n_bins / n_ffa_bins
 *    Each new subfreq is obtained from the previous one by multiplication
 *    with 1 + 1 / n_bins, which we can do about n_bins / subfreq times before
 *    we change n_per, i.e. int(subfreq), thereby changing n_ffa_bins.
 *
 *  2) Include all coadds.
 *  coadds = log_2(n_ffa_bins / min_freq) + 1
 *   (For n_ffa_bins = base_freq*2^n, we must count coadd 0 through n.  For instance,
 *    for n = 2, the given nfb will come from coadds of 3 subfreqs:)
 *    * * * * * * * *
 *    **  **  **  **
 *    ****    ****
 *
 *  3) Number of translations allowed  (phases of the wave)
 *  trans = n_bins / n_ffa_bins
 *   (The number of right and left translations of n_ffa_bins bins that fit into n_bins.
 *    In the example below, the number is 2:)
 *    **  **  **  **
 *      **  **  **  **
 *
 * After we've counted up a certain number of bins, the value of which is stored in
 * key_increment_bins, we store a key_nfb, and start counting bins from 0 again.
 */

void set_key_nfb(std::vector<int> &key_nfb, float min_freq, int n_bins) {
	int total_bins = 0;
	int bins_per_nfb_subfreq = 0;
	key_nfb.clear();

	/* key_increment_bins: once we've counted this number of lcgfs, move on to the next key_nfb */
	int key_increment_bins = (int)(((double)n_bins * (double)n_bins / min_freq) / 240);
	int test_val;

	for (int nfb_subfreq = (int)min_freq; nfb_subfreq < n_bins; nfb_subfreq++) {
	  bins_per_nfb_subfreq = (int)(((double)n_bins * n_bins) / ((double)nfb_subfreq * nfb_subfreq) * (int_log2((int)(nfb_subfreq/min_freq)) + 1));
	  total_bins += bins_per_nfb_subfreq;
	  if (total_bins >= key_increment_bins) {
		total_bins = 0;
		key_nfb.push_back(nfb_subfreq);
		for (std::vector<int>::iterator iter = key_nfb.begin(); iter != key_nfb.end(); iter++) test_val = *iter;
	  }
	}
	key_nfb.push_back(n_bins-1);  // One last key_nfb for the ones that don't have a place yet
} //set_key_nfb

//R:Returns true if repetitive pulse limit reached to speedup computations
inline bool LogRepetitivePulse(float temp_coadd[],int m,int temp_coadd_working_length,int coadd,
				   int n_client_bins,int dm_chunk,int fold_level,bool is_long,
				   int pulse_limit_single, int pulse_limit_rep,double thresh_x_intercept,
				   double thresh_x_slope,double period)
{
		//bin_count++;
			  /* If a bin goes over threshhold, record it.
			   * We determine if a bin is over threshhold by deciding whether its
			   * probability of existence (in random noise) is < 1/(number of bins we examine per workunit)
			   * This ensures that on average we will detect 1 false positive.
			   * There are N freq's, where (1 + 1/sub_buffer_size)^N = 2.
			   * That is, e^(N/sub_buffer_size) = 2, so N = sub_buffer_size * ln 2.
			   * Each freq has a period equal to sub_buffer_size / freq bins, for a
			   * total of about sub_buffer_size^2 / state.min_freq terms.
			   * And sub_buffer_size = num_samples / (dm_chunk * state.num_sub_buffers), so that makes
			   * num_samples^2 / (dm_chunk^2 * state.num_sub_buffers^2 * state.min_freq)
			   * Second, for each freq, there are a number of subfreqs, each of which has
			   * half as many bins as the previous.  This doubles the number of bins.
			   * Third, for each subfreqs, there are a number of coadds, each of which
			   * has half as many bins as the previous.  This doubles the number of
			   * bins again.
			   * Fourth, the code runs state.num_sub_buffers times, so we multiply by that.
			   * Finally, there are 15000 / dm_chunk dm chunks, and we run
			   * the ffa once per dm chunk.
			   * Therefore, the number of bins we examine per workunit is
			   * (num_samples^2 * 4 * 15000) / (state.num_sub_buffers * state.min_freq * dm_chunk^3).
			   * Currently, num_samples = 32 * 10^6, state.num_sub_buffers = 1, state.min_freq = 137,
			   * and dm_chunk = 128.
			   * That's 2 * 10^11 bins per workunit, which is e^26.18
			   * This suggests that the lcgf should have a threshold of -26.  But evidently this isn't enough ... ?
			   * Note: for dm_chunk_small, change sub_buffer_size to fold_buf_bytes_short * 4 / dm_chunk
			   * (fold_buf_bytes_short^2 * 16 * 4 * 15000) / (state.num_sub_buffers * state.min_freq
			   *                * dm_chunk^3)    ... I get e^(-25.5)
			   *
			   * First, total power in these bins must exceed the number of client bins.
			   */
			  //if(temp_coadd[m] > n_client_bins + 1) { R:Unneeded check.
			  /* If we're writing the thresholds file, then
			   * check if we have a new max thresh_x for this key_nfb.
			   */

#if 0
		  fprintf(stderr,"CPU rescan was needed indeed\n");
#endif

				  // Temporary split between if statements removed
				  //pulse_count++;
	  state_t& state = client.ap_shmem->ap_gdata.state;
	  ap_signal s;
				  time_series_b_t time_series;
				  time_series.build(temp_coadd, m, temp_coadd_working_length, (double) n_client_bins);

				  /* I don't understand why state.dm_now / dm_chunk is here. */
				  state.pulse_jd = state.jd0 + (state.data_chunk_now + state.dm_now/dm_chunk)
										  /(state.datasize)*(state.jd1-state.jd0);
				  s.time=state.pulse_jd;
				  // s.peak_bin = state.data_chunk_now*4 + m;
				  s.peak_bin = m * dm_chunk * (1 << coadd);
				  s.peak_power = temp_coadd[m];
				  s.scale = fold_level;
				  if(is_long)
					s.dm = state.dm_sign*(state.dm_chunk_large_now);
	  else
					s.dm = state.dm_sign*(state.dm_chunk_large_now+state.dm_chunk_small_now);
				  s.period = period;
				  s.ffa_scale = coadd;
				  // Machinery for writing s.num_std_devs.  fudge_factor is about 1.
				  float fudge_factor;
				  fudge_factor = 1 - thresh_x_slope * (n_client_bins>>fold_level) / thresh_x_intercept;
				  s.num_std_devs = (s.peak_power - n_client_bins)/(fudge_factor * sqrt((double)n_client_bins));
				  s.time_series=time_series.data;
				  if(add_signal(client.signal_vector, s)) {state.result_count_rep++;
#if 0
	  fprintf(stderr,"Signal added\n");
		 }else{
	  fprintf(stderr,"Signal discarded\n");
#endif
		 }
				  // client.outfile.output_result(s);
				  if(pulse_limit_rep && state.result_count_rep >= pulse_limit_rep)
					  {
					  // We've already found enough repeating pulses.
					  if(pulse_limit_single && state.result_count_single >= pulse_limit_single)
						 {
						 // Quit early
						 printf("Found %d single pulses and %d repeating pulses, exiting.\n", state.result_count_single, state.result_count_rep);
						 fprintf(stderr, "Found %d single pulses and %d repeating pulses, exiting.\n", state.result_count_single, state.result_count_rep);
						 fflush(stdout);
						 client.outfile.write_main(client.signal_vector);
						 client.outfile.finish_results();
						 client.finish();
						 main_timer.stop();
						 boinc_finish(0);
						 exit(0);
						 }
					  else
						 {
						 //(Must be that we haven't found enough single pulses.)
						 // Don't waste time searching for more repetitive pulses.
						 skipffa_short = true;                     // bypass shorts
						 skipffa_long = true;                      // bypass longs
						 state.fold_level_small = 99;              // inhibit short save
						 state.fold_level_large = 99;              // inhibit long save
						 return true;
						 }
					  }
		  return false;
}

} //end namespace Astropulse


