// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/* ap_science.C */
/* Functions for AstroPulse scientific computations */

#ifndef _WIN32
#include "ap_config.h"
#else
#include "boinc_win.h"
#include "intrinsics.h"
#endif

#include <cmath>
#include <cassert>
#if __FreeBSD__
#include <stdlib.h>
#endif
#include "astropulse.h"
#include "boinc_api.h"
#include "sbtf.h"
#include "ap_debug.h"
#if USE_FFTW
	#include "fftw3.h"
#endif
#if USE_IPP
	#include "ipp_static.h"
#endif
#if USE_ACML
	#include "acml.h"
	#include "fwbase.h"
#endif
/* Windows is stupid */
#ifndef M_PI
#define M_PI 3.141592653589793238462
#endif

#ifdef _MSC_VER
#define ALIGN16_BEG __declspec(align(16))
#define ALIGN16_END
#else /* gcc or icc */
#define ALIGN16_BEG
#define ALIGN16_END __attribute__((aligned(16)))
#endif

extern bool modev505;
extern APP_INIT_DATA app_init_data;

namespace Astropulse {
#if USE_FFTW
#include "fftw3.h"

inline void Science::Wisdom::load() {
	strcpy(wname,app_init_data.project_dir);
   #ifdef _WIN32
	strcat(wname, "\\");
   #else
	strcat(wname, "/");
   #endif
	strcat(wname,WISDOM_FNAME);
	FILE *wisdom=boinc_fopen(wname,"r");
	if (wisdom) { //uje: moved up
	  char *wiz=(char *)calloc(1024,64);
	  int n=0;
	  while (wiz && n<64*1024 && !feof(wisdom)) {
		n+=fread(wiz+n,1,80,wisdom);
	  }
	  fftwf_import_wisdom_from_string(wiz);
	  free(wiz);
	  fclose(wisdom);
	}
}

inline void Science::Wisdom::save() {
	// Save new wisdom
	FILE *wisdom=boinc_fopen(wname,"w");
	if (wisdom) {
	  char *wiz=fftwf_export_wisdom_to_string();
	  if (wiz) {
		fwrite(wiz,strlen(wiz),1,wisdom);
	  }
	  fclose(wisdom);
	}
}
#endif// USE_FFTW

// Allocate memory, and create fft plans
#if USE_IPP
	int Science::order=15;
#endif
//extern FILE* r_debug;
void Science::init() {
	state_t& state = client.ap_shmem->ap_gdata.state;
	const size_t& fft_len = state.fft_len;
	int fft_len2 = 1024;
	// New stuff for 5.01
	fftlen_envelope = 128;
	fftlen_patch = 512;

  //JWS: Moved from ap_remove_radar.cpp Science::generate_envelope()
  // Check whether envelope size is (wrongly) larger than fft len
  // This should never happen, but we check just in case.
  if(fftlen_envelope > state.fft_len) {
	printf("Error in ap_science.cpp : Science::init : fftlen_envelope is > client fft_len and should not be.\n");
	fprintf(stderr, "Error in ap_science.cpp : Science::init : fftlen_envelope is > client fft_len and should not be.\n");
  }

	//R: Some hack to reduce memory requirements and speedup memory copying
//  temp_buf_long_size=4096;//R:2048 is enough for stock WU but my shortened one requires this limit (near pow of 2 that exceeds actual value)
//  temp_buf_short_size=512;//R: actual value is 479
	const size_t& fold_buf_size_long=state.fold_buf_size_long;
	const size_t& fold_buf_size_short=state.fold_buf_size_short;

	//JWS: actual needed temp buffer sizes rounded up to 4 KiB page size ( 1K floats),
	// will continue working even with changed WU header parameters.
	temp_buf_long_size = ((unsigned)ceil((double)fold_buf_size_long / fabs(state.min_freq)) + 1024) & (~0x3FF);
	temp_buf_short_size = ((unsigned)ceil((double)fold_buf_size_short / fabs(state.min_freq)) + 1024) & (~0x3FF);

	//JWS: temp, temp_coadd, temp_o2, and temp_o3 buffer allocation for FFA.
	// 256 bytes (64 floats) is added to avoid untimely L1 cache line evictions.
	unsigned tbs = 64 + std::max(temp_buf_long_size, temp_buf_short_size);

	//R: Allocation of all big arrays done here. Both for dechirping, FFT and FFA.
	//R:data and chirp arrays will be used together so we need proper padding to exclude cache contentions.
	const int padding=512;//R:all elements still need 16-byte alignment so padding should be multiple of 4 (padding in numbers of floats)
	//R: temp_fold_buf_* could be much shorter, need to figure out how big they should be really.

	// Allocations phase 1: set mass app_data_array allocation if used.

#if _WIN32 && USE_VIRTUAL_ALLOC
	// app_data_array not used
//#elif _GNU_SOURCE // uncomment if virtual allocated memory replacement with gcc
	// app_data_array not used
#elif USE_GPU
	// Allocate CPU side app_data_array
	gpuError=cudaMallocHost((void**)&app_data_array,sizeof(float)*(fft_len*2*(2+state.dm_chunk_small)+fft_len/*filter*/+10*padding+
	  2*fftlen_patch+2*fftlen_envelope+2*(fold_buf_size_long+4)+2*(fold_buf_size_short+4)+2*tbs)+4096);
	if (!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error app_data_array, exiting...\n");
			exit(0);
	}
#else
	//JWS: summing allocations in floats
	unsigned asz = 1024;                   // allowance for initial page alignment
	asz += fft_len*2;                      // temp (or temp and i_temp for split)
	asz += fft_len*2;                      // data (or data and i_data for split)
	asz += fft_len*2;                      // filter_array
	asz += fft_len*2*state.dm_chunk_small; // chirp (or chirp and i_chirp for split)
	asz += fft_len*2+1024;                 // temp_in
  #if TWINDECHIRP
	asz += fft_len*2+1024;                 // temp_in_neg
  #endif
	asz += 2*(fold_buf_size_long+1024);    // fold_buf_long_pos and fold_buf_long_neg
	asz += 2*(fold_buf_size_short+1024);   // fold_buf_short_pos and fold_buf_short_neg
	asz += 4*tbs+1024;                     // temp_long, temp_coadd_long, temp_o2, temp_o3
	asz += 2*fftlen_patch;                 // patch buffer
	asz += 2*fftlen_patch;                 // patch_mults
	asz += 2*fftlen_envelope;              // envelope_part buffer

	app_data_array=(float*) calloc(asz, sizeof(float));

#endif

	// Allocations Phase 2: Do separate allocations or set pointers into mass allocation.

#if _WIN32 && USE_VIRTUAL_ALLOC

	#define VAlloc_f(_size) (float *) VirtualAlloc(NULL,_size,MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE )
	#define VAlloc_d(_size) (double *) VirtualAlloc(NULL,_size,MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE )
	#define VAlloc_c(_size) (ap_complex *) VirtualAlloc(NULL,_size,MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE )
	#define VFree(_ptr) VirtualFree(_ptr,0,MEM_RELEASE)

	temp = VAlloc_c(sizeof(fftwf_complex)*fft_len);
	data = VAlloc_c(sizeof(fftwf_complex)*fft_len);
	filter_array = VAlloc_f(sizeof(fftwf_complex)*fft_len);
	chirp = VAlloc_c(sizeof(fftwf_complex)*fft_len*state.dm_chunk_small);
	temp_in_allo = VAlloc_c(sizeof(fftwf_complex)*fft_len+4096);
	temp_in = temp_in_allo+384; // working part at 3/4 page
  #if TWINDECHIRP
	temp_in_neg_allo = VAlloc_c(sizeof(fftwf_complex)*fft_len+4096);
	temp_in_neg = temp_in_neg_allo+384; // working part at 3/4 page
	negredy = -1;
  #endif
	fold_buf_long_pos = VAlloc_f(sizeof(float)*(fold_buf_size_long +1024) );
	fold_buf_long_neg = VAlloc_f(sizeof(float)*(fold_buf_size_long +1024) );
	fold_buf_short_pos = VAlloc_f(sizeof(float)*(fold_buf_size_short +1024) );
	fold_buf_short_neg = VAlloc_f(sizeof(float)*(fold_buf_size_short +1024) );
	temp_long = VAlloc_f(sizeof(float)*tbs*4);
	temp_coadd_long = 	temp_long+tbs;
	temp_o2 = temp_coadd_long+tbs;
	temp_o3 = temp_o2+tbs;
	temp_short = temp_long;
	temp_coadd_short = temp_coadd_long;
	patch = VAlloc_c(sizeof(fftwf_complex)*fftlen_patch);
	patch_mults = VAlloc_d(sizeof(double)*fftlen_patch);
	envelope_part = VAlloc_c(sizeof(fftwf_complex)*fftlen_envelope);

/*
#elif _GNU_SOURCE
	//uje: Spot for virtual allocated memory replacement with gcc
	data = (ap_complex*)realloc(data, sizeof(fftwf_complex)*fft_len);
	temp  = (ap_complex*)realloc(temp, sizeof(fftwf_complex)*fft_len);
	chirp  = (ap_complex*)realloc(chirp, sizeof(fftwf_complex)*fft_len*state.dm_chunk_small);
	patch = (ap_complex*)malloc(sizeof(fftwf_complex)*fftlen_patch);
	envelope_part = (ap_complex*)malloc(sizeof(fftwf_complex)*fftlen_envelope);
	filter_array = (float*)malloc(sizeof(fftwf_complex)*fft_len);
	fold_buf_long_pos = (float*)malloc(sizeof(float)*(fold_buf_size_long +4) );
	fold_buf_long_neg = (float*)malloc(sizeof(float)*(fold_buf_size_long +4) );
	fold_buf_short_pos = (float*)malloc(sizeof(float)*(fold_buf_size_short +4) );
	fold_buf_short_neg = (float*)malloc(sizeof(float)*(fold_buf_size_short +4) );
	temp_long = (float*)malloc(sizeof(float)*tbs );
	temp_coadd_long = (float*)malloc(sizeof(float)*tbs );
	temp_short = temp_long;
	temp_coadd_short = temp_coadd_long;
*/

#elif ( __SSE__ && SPLIT_COMPLEX)
	temp=(float*)((unsigned long)app_data_array&(~0xFFF))+0x400;//temp[] (and power[] later) aligned on page boundary
	/*R: & JWS:
	data arrangement:
	temp[]=32k*4,padding,i_temp=32k*4,padding,data=32k*4,padding,i_data=32k*4,padding,
	chirp=32k*4,padding,i_chirp=32k*4,filter_array,padding,patch,envelope_part,
	fold_buf_long_pos,fold_buf_long_neg,temp_fold_buf_long,temp_coadd_long, // also used for short
	fold_buf_short_pos,fold_buf_short_neg
	*/
	i_temp=temp+fft_len+padding;
	data=i_temp+fft_len + padding;
	i_data=data+fft_len + padding;
	chirp=i_data+fft_len + padding;
	i_chirp=chirp+fft_len*state.dm_chunk_small+padding;
	filter_array=i_chirp+fft_len*state.dm_chunk_small+padding;
	patch=(ap_complex*)(filter_array+fft_len+padding);
	envelope_part=patch+fftlen_patch;//JWS: increment in ap_complex units
	fold_buf_long_pos=(float*)(envelope_part)+2*fftlen_envelope;
	fold_buf_long_neg=fold_buf_long_pos+fold_buf_size_long+4;//R:Alloc one additional trash element to avoid if(l==n_bins) in inner loop of ffa()
	temp_long=fold_buf_long_neg+(fold_buf_size_long+4);//R:Alloc one additional trash element to avoid if(l==n_bins) in inner loop of ffa()
	temp_coadd_long=temp_long+tbs;
	fold_buf_short_pos=temp_coadd_long+tbs;
	fold_buf_short_neg=fold_buf_short_pos+(fold_buf_size_short+4);//R:Alloc one additional trash element to avoid if(l==n_bins) in inner loop of ffa()
	temp_short=temp_long;
	temp_coadd_short=temp_coadd_long;

#else  // Not split_complex, not separate allocations
	temp=(ap_complex*)((unsigned long)app_data_array&(~0xFFF))+0x200;//temp[] (and power[] later) aligned on page boundary
	data=(ap_complex*)((float*)temp+2*fft_len); // also aligned
	filter_array=(float*)((float*)data+2*fft_len); // also aligned
	chirp=(ap_complex*)(filter_array+fft_len); // also aligned
	temp_in=(ap_complex*)((float*)chirp+2*fft_len*state.dm_chunk_small+768); // at 3/4 page
  #if TWINDECHIRP
	temp_in_neg=(ap_complex*)((float*)temp_in+2*fft_len); // also at 3/4 page
	negredy = -1;
	fold_buf_long_pos=(float*)temp_in_neg+2*fft_len+256; // page aligned
  #else
	fold_buf_long_pos=(float*)temp_in+2*fft_len+256; // page aligned
  #endif
	fold_buf_long_neg=fold_buf_long_pos+fold_buf_size_long+1024; // page aligned
	fold_buf_short_pos=fold_buf_long_neg+fold_buf_size_long+1024; // page aligned
	fold_buf_short_neg=fold_buf_short_pos+fold_buf_size_short+1024; // page aligned
	temp_long=fold_buf_short_neg+fold_buf_size_short+1024; // page aligned
	temp_coadd_long=temp_long+tbs; // page + 64
	temp_o2=temp_coadd_long+tbs; // page + 128
	temp_o3=temp_o2+tbs; // page + 192
	patch=(ap_complex*)((float*)temp_o3+tbs); // page + 256
	patch_mults=(double*)((float*)patch+2*fftlen_patch);
	envelope_part=(ap_complex*)((float*)patch_mults+2*fftlen_patch);
	temp_short=temp_long;
	temp_coadd_short=temp_coadd_long;
	if ((app_data_array+asz) < ((float*)envelope_part+2*fftlen_envelope))
		  fprintf(stderr, "\n*** Mass  allocation not enough! ***\n\n");
#endif

//Memory alloc on CUDA GPU
#if USE_GPU
	gpuError=cudaMalloc((void**)&gpudata,sizeof(cufftComplex)*fft_len);//GPU array for FFT by CUFFT - D: this array MUST be in GPU memory ...
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpudata, exiting...\n");
			exit(0);
	}
	gpuError=cudaMalloc((void**)&gputemp,sizeof(cufftComplex)*fft_len);//GPU array D: this array MUST be in GPU memory ...
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gputemp, exiting...\n");
			exit(0);
	}
	gpuError=cudaMalloc((void**)&gpupower,sizeof(float)*fft_len/2);//GPU array D: this array MUST be in GPU memory ...
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpupower, exiting...\n");
			exit(0);
	}
	gpuError=cudaMalloc((void**)&gpuchirp,sizeof(cufftComplex)*fft_len*state.dm_chunk_small);//GPU array for FFT by CUFFT - D: this array MUST be in GPU memory ...
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpuchirp, exiting...\n");
			exit(0);
	}
	gpuError=cudaMalloc((void**)&gpu_filter_array,sizeof(float2)*fft_len);//R:for high_pass_filter() on GPU
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpu_filter_array, exiting...\n");
			exit(0);
	}
	gpuError=cudaMalloc((void**)&gpu_fold_buf_short_pos,sizeof(float)*(fold_buf_size_short+4));//R:for FFA on GPU
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpu_fold_buf_short_pos, exiting...\n");
			exit(0);
	}
	gpuError=cudaMalloc((void**)&gpu_fold_buf_short_neg,sizeof(float)*(fold_buf_size_short+4));//R:for FFA on GPU
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpu_fold_buf_short_neg, exiting...\n");
			exit(0);
	}
	gpuError=cudaMalloc((void**)&gpu_temp_buf_short,sizeof(float)*(tbs+4));
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpu_temp_buf, exiting...\n");
			exit(0);
	}
	gpu_temp_buf_long = gpu_temp_buf_short;

	gpuError=cudaMalloc((void**)&gpu_temp_coadd_short,sizeof(float)*(tbs+4));
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpu_temp_coadd, exiting...\n");
			exit(0);
	}
	gpu_temp_coadd_long = gpu_temp_coadd_short;

	gpuError=cudaMalloc((void**)&gpu_fold_buf_long_pos,sizeof(float)*(fold_buf_size_long+4));//R:for FFA on GPU
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpu_fold_buf_long_pos, exiting...\n");
			exit(0);
	}
	gpuError=cudaMalloc((void**)&gpu_fold_buf_long_neg,sizeof(float)*(fold_buf_size_long+4));//R:for FFA on GPU
	if(!gpuError==0){
	  fprintf(stderr,"ERROR: data alloc error gpu_fold_buf_long_neg, exiting...\n");
			exit(0);
	}
#else
/*JWS: not needed for release builds
  #if USE_VIRTUAL_ALLOC
	fprintf(stderr, "VirtualAlloc:\n");
  #else
	fprintf(stderr, "Mass allocation:  app_data_array=0x%X\n", app_data_array);
  #endif
	fprintf(stderr, "temp=0x%X  data=0x%X  filter_array=0x%X  chirp=0x%X\n",
			temp, data, filter_array, chirp);
	fprintf(stderr, "temp_in=0x%X  fold_buf_long_pos=0x%X  fold_buf_long_neg=0x%X\n",
			temp_in, fold_buf_long_pos, fold_buf_long_neg);
	fprintf(stderr, "fold_buf_short_pos=0x%X  fold_buf_short_neg=0x%X  temp_long=0x%X\n",
			fold_buf_short_pos, fold_buf_short_neg, temp_long);
	fprintf(stderr, "temp_coadd_long=0x%X  temp_o2=0x%X  temp_o3=0x%X  patch=0x%X\n",
			temp_coadd_long, temp_o2, temp_o3, patch);
	fprintf(stderr, "patch_mults=0x%X  envelope_part=0x%X  ", patch_mults, envelope_part);
  #if TWINDECHIRP
	fprintf(stderr, "temp_in_neg=0x%X\n\n", temp_in_neg);
  #else
	fprintf(stderr, "\n\n");
  #endif
*/
#endif

#if BOINC_APP_GRAPHICS
	client.status = "Creating FFT plans.";
#endif

#ifndef BOINC_APP_GRAPHICS
	client.power=(float*)temp;
#endif

#if USE_GPU //FFT plan calculation for CUFFT
	gpufftError=cufftPlan1d(&cuFFTplan,fft_len,CUFFT_C2C,1); // D: plan generation for CUFFT
	if (!gpufftError==0){
	  fprintf(stderr,"ERROR: fft plan error\n");
	exit(-1);
	}
#endif

#if USE_FFTW
	wisdom.load();
  #if ( __SSE__ && SPLIT_COMPLEX) //uje: && !_GNU_SOURCE
	fftw_iodim fft_dim;
	fft_dim.n=fft_len;
	fft_dim.is=fft_dim.os=1;
	fp = fftwf_plan_guru_split_dft(1,&fft_dim,0,NULL,data, i_data, data,i_data, FFTW_MEASURE);
	ip = fftwf_plan_guru_split_dft(1,&fft_dim,0,NULL,i_temp, temp, i_temp, temp, FFTW_MEASURE);
	fp2 = fftwf_plan_dft_1d(fft_len2, data, data, FFTW_FORWARD, FFTW_MEASURE);
	#ifndef BLANKIT
	fp_patch_inv = fftwf_plan_dft_1d(fftlen_patch, patch, patch, FFTW_BACKWARD, FFTW_MEASURE);
	fp_envelope = fftwf_plan_dft_1d(fftlen_envelope, envelope_part, envelope_part, FFTW_FORWARD, FFTW_MEASURE);
	#endif

/*#elif _GNU_SOURCE && ( __SSE__ && SPLIT_COMPLEX) //uje: for later trying to make SPLIT_COMPLEX work
 fftw_iodim fft_dim;
 fft_dim.n=fft_len;
 fft_dim.is=fft_dim.os=1;
 fp = fftwf_plan_guru_split_dft(1,&fft_dim,0,NULL,data, i_data, data,i_data, FFTW_MEASURE);
 ip = fftwf_plan_guru_split_dft(1,&fft_dim,0,NULL,i_temp, temp, i_temp, temp, FFTW_MEASURE);
 fp2 = fftwf_plan_dft_1d(fft_len2, data, data, FFTW_FORWARD, FFTW_MEASURE);
#ifndef BLANKIT
 fp_patch_inv = fftwf_plan_dft_1d(fftlen_patch, patch, patch, FFTW_BACKWARD, FFTW_MEASURE);
 fp_envelope = fftwf_plan_dft_1d(fftlen_envelope, envelope_part, envelope_part, FFTW_FORWARD, FFTW_MEASURE);
#endif
*/

  #else
	fp2 = fftwf_plan_dft_1d(fft_len2, data, data, FFTW_FORWARD, FFTW_MEASURE);
	#ifndef BLANKIT
	fp_patch_inv = fftwf_plan_dft_1d(fftlen_patch, patch, patch, FFTW_BACKWARD, FFTW_MEASURE);
	fp_envelope = fftwf_plan_dft_1d(fftlen_envelope, envelope_part, envelope_part, FFTW_FORWARD, FFTW_MEASURE);
	#endif
	//JWS: Doing out-of-place for both USE_VIRTUAL_ALLOC and mass allocation
	ip = fftwf_plan_dft_1d(fft_len, temp_in, temp, FFTW_BACKWARD, FFTW_PATIENT);
	fp = fftwf_plan_dft_1d(fft_len, temp_in, data, FFTW_FORWARD, FFTW_MEASURE);
	#if TWINDECHIRP
	ipn = fftwf_plan_dft_1d(fft_len, temp_in_neg, temp, FFTW_BACKWARD, FFTW_PATIENT);
	#endif
  #endif // ( __SSE__ && SPLIT_COMPLEX)
	wisdom.save();
#endif //USE_FFTW

#if USE_ACML //by Devaster
	fp=(complex *)malloc((5*fft_len+100)*sizeof(complex));
	fp2=(complex *)malloc((5*fft_len2+100)*sizeof(complex));
  #ifndef BLANKIT
	fp_patch_inv=(complex *)malloc((5*fftlen_patch+100)*sizeof(complex));
	fp_envelope=(complex *)malloc((5*fftlen_envelope+100)*sizeof(complex));
  #endif
	ip=(complex *)malloc((5*fft_len+100)*sizeof(complex));
	cfft1d(100,fft_len,(complex *)data,fp,&info);
	cfft1d(100,fft_len2,(complex *)data,fp2,&info);
	cfft1d(100,fftlen_patch,(complex *)patch,fp_patch_inv,&info);
	cfft1d(100,fftlen_envelope,(complex *)envelope_part,fp_envelope,&info);
	cfft1d(100,fft_len,(complex *)temp,ip,&info);
	acmlfftlen=fft_len;
	acmlfftlen2=fft_len2;
	acmlfftlenpatch=fftlen_patch;
	acmlfftlenenv=fftlen_envelope;
#endif //USE_ACML

//R:Extending JDWhale's caching. multipliers array can be calculated only once per app run.
	{//R: initialization of filter array
	const size_t& fft_len1 = state.fft_len;
	float sigma = client.ap_shmem->ap_gdata.state.high_pass * ((float) fft_len1 / 2500000);
	float rcp_2SigmaSqr = 1.0 / (2 * sigma * sigma);
	float normalize = 1 / sqrt(1 - sqrt(2 * 3.1415926) * sigma / fft_len1);
	float Gaussian, multiplier;
#if 0

	fprintf(stderr,"Filter array init\n");
#endif
	for(unsigned k=0;k<fft_len1/2;k++){
		Gaussian = exp(-((k+0.5f) *(k+0.5f) *rcp_2SigmaSqr));
		multiplier = sqrt(1 - Gaussian);
		multiplier *= normalize;
		filter_array[2*k]=multiplier;
		filter_array[2*k+1]=multiplier;//R:for easier SIMD transition
	}
#if 0
	fprintf(stderr,"Building completed, parameter values are:\n");
	fprintf(stderr,"fft_len=%u\tsigma=%e\trcp_2SigmaSqr=%e\tnormalize=%e\n",fft_len1,sigma,rcp_2SigmaSqr,normalize);
	for(unsigned k=0;k<fft_len1/2;k++){
		fprintf(stderr,"multiplier[%u]=%e\n",k,filter_array[2*k]);
	}
	fprintf(stderr, "Dump completed, exiting...\n");
	exit(0);
#endif
#if USE_GPU
	gpuError=cudaMemcpy(gpu_filter_array,filter_array,sizeof(float2)*fft_len,cudaMemcpyHostToDevice);
	if (!gpuError==0)
		{
			fprintf(stderr,"GPU memory copy error (data up) ...");
			exit(0);
		}
#endif
	}//R: end of filter init
} // end Science::init()

void Science::finish() {
#if USE_GPU
	cudaFree(gpudata);
	cudaFree(gputemp);
	cudaFree(gpupower);
	cudaFree(gpuchirp);
	cudaFree(gpu_fold_buf_short_pos);
	cudaFree(gpu_fold_buf_short_neg);
	cudaFree(gpu_temp_buf_short);
	cudaFree(gpu_temp_coadd_buf_short);
	cudaFree(gpu_fold_buf_long_pos);
	cudaFree(gpu_fold_buf_long_neg);
	cudaFreeHost(app_data_array);

#elif _WIN32 && USE_VIRTUAL_ALLOC
	//    VirtualFree(app_data_array,0,MEM_RELEASE);
	VFree(data);
	VFree(temp);
	VFree(temp_in_allo);
#if TWINDECHIRP
	VFree(temp_in_neg_allo);
#endif
	VFree(chirp);
	VFree(patch);
	VFree(patch_mults);
	VFree(envelope_part);
	VFree(filter_array);
	VFree(fold_buf_long_pos);
	VFree(fold_buf_long_neg);
	VFree(fold_buf_short_pos);
	VFree(fold_buf_short_neg);
	VFree(temp_long);

/*#elif _GNU_SOURCE //uje: for later trying to make SPLIT_COMPLEX work
	free(data);
	free(temp);
	free(chirp);
	free(patch);
	free(envelope_part);
	free(filter_array);
	free(fold_buf_long_pos);
	free(fold_buf_long_neg);
	free(fold_buf_short_pos);
	free(fold_buf_short_neg);
	free(temp_long);
	free(temp_coadd_long);*/

#else
	free(app_data_array);
#endif //USE_GPU

#if USE_FFTW
	fftwf_destroy_plan(fp);
	fftwf_destroy_plan(fp2);
	fftwf_destroy_plan(ip);
  #if TWINDECHIRP
	fftwf_destroy_plan(ipn);
  #endif
	fftwf_destroy_plan(fp_patch_inv);
#elif USE_IPP
	ippsFFTFree_C_32fc(pFFTSpec);
	//ippsFree(chirp);
	//ippsFree(temp);
	//ippsFree(data);
#endif
}

/* Science::build_chirp_table
 *
 * arguments:
 * int dm_start: the first dm to use in the chirp table.
 *  The chirp table will use all dms from dm_start to
 * dm_start + client.ap_shmem->ap_gdata.state.dm_chunk_small - 1.
 *
 * This function creates an array of complex numbers
 * (of type fftw_complex) which
 * represent a chirp.  The array indices represent frequencies, so
 * the chirp is described in frequency space.  We will multiply the
 * chirp by the frequency-space representation of the incoming signal
 * in order to dechirp the signal, so it is really a "dechirp" table.
 * That is, we use the reciprocal of the chirp, instead of the chirp
 * itself.  Since the chirp's amplitude is 1 everywhere, taking its
 * reciprocal is a simple matter of taking its complex conjugate.
 *
 * The first <state.fft_len> complex numbers correspond to the
 * dm indicated by dm_start.  The next set of numbers correspond to
 * dm_start+1, and so on until dm_start + client.ap_shmem->ap_gdata.state.dm_chunk_small - 1.
 */
void Science::build_chirp_table(int dm_start) {
	int i,j;
	double freq;
	double phase;
#if BOINC_APP_GRAPHICS
	client.status = "Building chirp table.";
#endif
	state_t& state = client.ap_shmem->ap_gdata.state;

	/* For debugging, see below
	static int idum = -1;
	*/

	// phase function determined by method of steepest descent.
	// 0.00176 = 2.5 MHz / 1.42 GHz.  When multiplied by the maximum
	// value freq = 0.5, it will give 1.25 MHz / 1.42 GHz.
//  for (i=0; i<state.dm_chunk_large; i++) {
	//JWS: Change to smaller table
	for (i=0; i<state.dm_chunk_small; i++) {
#if ( __SSE__ && SPLIT_COMPLEX) //uje: && !_GNU_SOURCE //>>>>>>>>>>> R: float phase instead double. Can lead to invalid results!!!
	  _m128 rnum=_mm_set1_ps(1.0/state.fft_len); //>>> R:maybe to use fixed value later
	  _m128 one=_mm_set1_ps(1.f);
	  _m128 freq_const=_mm_set1_ps(0.00176f);
	  for(j=0; j<state.fft_len;j+=4){
		static const ALIGN16_BEG float _ps_0123[4] ALIGN16_END = { 0.f, 1.f, 2.f, 3.f };
		_m128 index=_mm_add_ps(_mm_set1_ps(j),*(_m128*)_ps_0123);
		_m128 freq=_mm_mul_ps(index,rnum);
		if(j>state.fft_len/2){freq=_mm_sub_ps(freq,one);}
		_m128 rfreq=_mm_rcp_ps(_mm_add_ps(one,_mm_mul_ps(freq_const,freq)));
		_m128 phase=_mm_mul_ps(rfreq,_mm_mul_ps(freq,freq));
		_m128 cos_chirp=_mm_load_ps(chirp+j);
		_m128 sin_chirp=_mm_load_ps(i_chirp+j);
		sincos_ps(phase,&sin_chirp,&cos_chirp);
		_mm_stream_ps(chirp+j,cos_chirp);
		_mm_stream_ps(i_chirp+j,sin_chirp);
	  }

/*#elif _GNU_SOURCE && ( __SSE__ && SPLIT_COMPLEX) //uje: for later trying to make SPLIT_COMPLEX work
	  __m128 rnum=_mm_set1_ps(1.0/state.fft_len);
	  __m128 one=_mm_set1_ps(1.f);
	  __m128 freq_const=_mm_set1_ps(0.00176f);
	  for(j=0; j<state.fft_len;j+=4){
		static const ALIGN16_BEG float _ps_0123[4] ALIGN16_END = { 0.f, 1.f, 2.f, 3.f };
		__m128 index=_mm_add_ps(_mm_set1_ps(j),*(__m128*)_ps_0123);
		__m128 freq=_mm_mul_ps(index,rnum);
		if(j>state.fft_len/2) freq=_mm_sub_ps(freq,one);
		__m128 rfreq=_mm_rcp_ps(_mm_add_ps(one,_mm_mul_ps(freq_const,freq)));
		__m128 phase=_mm_mul_ps(rfreq,_mm_mul_ps(freq,freq));
		__m128 cos_chirp=_mm_load_ps(chirp+j);
		__m128 sin_chirp=_mm_load_ps(i_chirp+j);
		sincos_ps(phase,&sin_chirp,&cos_chirp);
		_mm_stream_ps(chirp+j,cos_chirp);
		_mm_stream_ps(i_chirp+j,sin_chirp);
	  }*/

#else
	  register const double phase_const=M_PI*(double)(dm_start+i);
	  for (j=0; j<state.fft_len; j++) {
	  // To get exact chirp function, we've multiplied the phase
	  // by 1 / (1 + 0.00176 * freq) from the linear chirp function
	  // This is equal to nu0 / nu, where nu0 = 1.42 GHz and
	  // nu = 1.42 GHz + freq in Hz = 1.42 GHz + (freq / (4 * 10^-7)) Hz,
	  // resulting in the value 0.00176 = 1 / (1.42 * 10^9 * 4 * 10^-7)
		freq = (double)j/(double)state.fft_len;
		freq = (j<state.fft_len/2)?freq:freq-1.0;
		phase = phase_const*freq*freq*(1 / (1 + 0.00176 * freq));

#if ( __SSE__ && SPLIT_COMPLEX) //uje: && !_GNU_SOURCE
	  chirp[i*state.fft_len + j] = (float)cos(phase);
	  i_chirp[i*state.fft_len + j] = (float)sin(phase);

/*#elif _GNU_SOURCE && ( __SSE__ && SPLIT_COMPLEX) //uje: for later trying to make SPLIT_COMPLEX work
	  chirp[i*state.fft_len + j] = (float)cos(phase);
	  i_chirp[i*state.fft_len + j] = (float)sin(phase);*/

#else
	  if (j<state.fft_len/2)
	  {
		 chirp[i*state.fft_len + j][0] = (float)cos(phase) * filter_array[2*j];
		 chirp[i*state.fft_len + j][1] = (float)sin(phase) * filter_array[2*j+1];
	  }
	  else
	  {
		 chirp[i*state.fft_len + j][0] = (float)cos(phase) * filter_array[2*state.fft_len-2*j-2];
		 chirp[i*state.fft_len + j][1] = (float)sin(phase) * filter_array[2*state.fft_len-2*j-1];
	  }
#endif
	}
#endif
  }

#if USE_GPU
 gpuError=cudaMemcpy(gpuchirp,chirp,sizeof(cufftComplex)*state.fft_len*client.ap_shmem->ap_gdata.state.dm_chunk_small,cudaMemcpyHostToDevice);
		if (!gpuError==cudaSuccess)
		{
			fprintf(stderr,"GPU memory copy error (data up) ...");
			exit(0);
		}
#endif
}

void Science::convert_bits_to_float() {
#if BOINC_APP_GRAPHICS
	client.status = "Converting raw data.";
#endif

#if ( __SSE__ && SPLIT_COMPLEX) //uje: && !_GNU_SOURCE
	if (splitter_bits_to_float(&client.rawdata[client.ap_shmem->ap_gdata.state.data_chunk_now], data, i_data, client.ap_shmem->ap_gdata.state.fft_len)) {

/*
#elif _GNU_SOURCE && ( __SSE__ && SPLIT_COMPLEX) //uje: for later trying to make SPLIT_COMPLEX work
	if (splitter_bits_to_float(&client.rawdata[client.ap_shmem->ap_gdata.state.data_chunk_now], data, i_data, client.ap_shmem->ap_gdata.state.fft_len)) {*/

#elif USE_FFTW // was _WIN32 && USE_VIRTUAL_ALLOC
	if (splitter_bits_to_float(&client.rawdata[client.ap_shmem->ap_gdata.state.data_chunk_now], (float*)temp_in, client.ap_shmem->ap_gdata.state.fft_len)) {
#else
	if (splitter_bits_to_float(&client.rawdata[client.ap_shmem->ap_gdata.state.data_chunk_now], (float*)data, client.ap_shmem->ap_gdata.state.fft_len)) {
#endif
	exit(SBTF_ERR);
  }
}

void Science::compute_forward_fft() {
#if !(( _GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
//Timings<T_FFT_forward> timings;
#endif

#if BOINC_APP_GRAPHICS
  client.status = "Computing forward FFT.";
#endif

#if 0
  //R: For debug purposes only. This block will generate harmonic
//signal in data[] to be passed to FFT.
  {int period_divider=16000;
fprintf(stderr,"DEBUG: Harmonic signal substitution enabled\n");
  for(int i=0;i<client.ap_shmem->ap_gdata.state.fft_len;i++){
	  data[i][0]=sin(period_divider*2*M_PI*i/(float)client.ap_shmem->ap_gdata.state.fft_len);
	  data[i][1]=0;
  }
  }
#endif
#if 0
//Dumping FFT input
  fprintf(stderr,"Before forward FFT\n");
  for(int i=0;i<client.ap_shmem->ap_gdata.state.fft_len;i++){
	  fprintf(stderr,"data[%d].re=%f,data[%d].im=%f\n",i,data[i][0],i,data[i][1]);
  }
#endif
#if USE_GPU
  {
  Timings<T_GPU_data_put> timings;
  gpuError=cudaMemcpy(gpudata,data,sizeof(ap_complex)*client.ap_shmem->ap_gdata.state.fft_len,cudaMemcpyHostToDevice);
	if (!gpuError==0)
		{
			fprintf(stderr,"GPU memory copy error (data up) ...");
			exit(0);
		}
  }

  gpufftError=cufftExecC2C(cuFFTplan,gpudata,gpudata,CUFFT_FORWARD);
		if (!gpufftError==0)
		{
			fprintf(stderr,"GPU FFT error ...");
			exit(0);
		}
#if 0
  {Timings<T_GPU_data_get> timings;
  gpuError=cudaMemcpy(data,gpudata,sizeof(ap_complex)*client.ap_shmem->ap_gdata.state.fft_len,cudaMemcpyDeviceToHost);
		if (!gpuError==0)
		{
			fprintf(stderr,"GPU memory copy error data<-gpudata ...");
			exit(0);
		}
  }
#endif

#elif USE_FFTW
  fftwf_execute(fp);
#elif USE_IPP
  ippsFFTFwd_CToC_32fc_I((Ipp32fc*)data,pFFTSpec,NULL); //R: maybe it's worth to provide some buffer for FFT...
#elif USE_ACML
  cfft1d(-1,acmlfftlen,(complex *)data,fp,&info);
#else
   #error "Some FFT lib needed"
#endif

#if 0//Dumping FFT output
  fprintf(stderr,"After forward FFT\n");
  for(int i=0;i<client.ap_shmem->ap_gdata.state.fft_len;i++){
	  fprintf(stderr,"data[%d].re=\t%e\tdata[%d].im=\t%e\n",i,data[i][0],i,data[i][1]);
  }
  exit(0);//stop execution for debug analysis.
#endif

}
#if 0
void Science::high_pass_filter() {
	Timings<T_filter> timings;
  /* One frequency bin is 2.5 MHz / fft_len worth of frequency.
   * So the width of the Gaussian filter in bins is:
   * (Note the sample rate is stored in the recorder_config, which is not part of the
   *  client.ap_shmem->ap_gdata.state.)
   */
	//const size_t& fft_len1 = client.ap_shmem->ap_gdata.state.fft_len;
#if 0
	//R:Set whole data array to 1 to see how filter array parforms in CPU/GPU. Only for debug.
	for(unsigned i=0;i<fft_len1;i++){
		data[i][0]=1;data[i][1]=1;
	}
	#if USE_GPU
		  gpuError=cudaMemcpy(gpudata,data,sizeof(ap_complex)*client.ap_shmem->ap_gdata.state.fft_len,cudaMemcpyHostToDevice);
		if (!gpuError==0)
		{
			fprintf(stderr,"GPU memory copy error data->gpudata ...");
			exit(0);
		}
#endif
#endif
  }
}
#else
void Science::high_pass_filter() {
	//Timings<T_filter> timings;
  /* One frequency bin is 2.5 MHz / fft_len worth of frequency.
   * So the width of the Gaussian filter in bins is:
   * (Note the sample rate is stored in the recorder_config, which is not part of the
   *  client.ap_shmem->ap_gdata.state.)
   */
#if USE_GPU
#if 0
//R: for the case if GPU memory can't last too long to survive between filter calls.
  gpuError=cudaMemcpy(gpu_filter_array,filter_array,sizeof(float)*client.ap_shmem->ap_gdata.state.fft_len,cudaMemcpyHostToDevice);
	if (!gpuError==0)
		{
			fprintf(stderr,"GPU memory copy error (data up) ...");
			exit(0);
		}
#endif
	GPUhigh_pass_filter(fft_len1,(float2*)gpudata,(float2*)gpu_filter_array);
#if 1
  {Timings<T_GPU_data_get> timings;
  gpuError=cudaMemcpy(data,gpudata,sizeof(ap_complex)*client.ap_shmem->ap_gdata.state.fft_len,cudaMemcpyDeviceToHost);
		if (!gpuError==0)
		{
			fprintf(stderr,"GPU memory copy error data<-gpudata ...");
			exit(0);
		}
  }
#endif

#else
	unsigned i;
	const size_t& fft_len1 = client.ap_shmem->ap_gdata.state.fft_len;
#if __SSE__
/*R:Further caching extension. The same multiplier used for Re and Im parts of data[i]. To avoid shuffling it's worth to dublicate multipliers array
at time of its creation. Then we can perform just streamlined SIMD multiplication w/o any shuffling.
*/
//Jason: base SIMD code ready to next integrate reverse filter (2nd half into first)...
//.. Then stripe in memory pages
	 for(i=0;i<fft_len1;i+=4) {   // First half is fft_len1/2 complexes = fft_len1 floats
	_mm_store_ps((float*)data+i, _mm_mul_ps(_mm_load_ps((float*)data+i), _mm_load_ps(filter_array+i)));
	}
//Second half: another fft_len1 floats, reverse filter (integrate into above later)
	 for(;i<2*fft_len1;i+=4)  {   // Second half is fft_len1/2 complexes = fft_len1 floats
	__m128 filter =  _mm_load_ps(filter_array+2*fft_len1-i-4);

	_mm_store_ps((float*)data+i, _mm_mul_ps(_mm_load_ps((float*)data+i), _mm_shuffle_ps( filter,filter,_MM_SHUFFLE(0,1,2,3))));
	}
#else
	for(i=0;i<fft_len1/2;i++){
		data[i][0]*=filter_array[2*i];
		data[i][1]*=filter_array[2*i];
	}
	for(;i<fft_len1;i++){
		data[i][0]*=filter_array[2*(fft_len1-1- i)];
		data[i][1]*=filter_array[2*(fft_len1-1- i)];
	}
#endif
#if 0
	int nf;  // For negative frequencies
	for (i = 0; i < (int) fft_len; i++) {
	  if (i < fft_len/2) {
		Gaussian = exp(-i * i *rcp_2SigmaSqr);
	  } else {
		nf = fft_len-1- i;
		Gaussian = exp(-nf * nf *rcp_2SigmaSqr);
	  }
	  /* Multiply the power by 1 - Gaussian. */
	  multiplier = sqrt(1 - Gaussian);
	  /* Conserve total noise power. */
	  multiplier *= normalize;
#if ( __SSE__ && SPLIT_COMPLEX)
			data[i] *= multiplier; //>>>>>>>> Need add true SIMD code
	  i_data[i] *= multiplier;
#else
	  data[i][0] *= multiplier;
	  data[i][1] *= multiplier;
#if 0
	  fprintf(stderr,"data[%d][0]=%e\tmultiplier=%e\n",i,data[i][0],multiplier);
#endif
#endif
	}
#endif
#endif
#if 0 //R:debug: dumping filtered array
	static int run_i;
	fprintf(stderr,"%d filter run finished @%.2f percent.\n",run_i++,client.ap_shmem->ap_gdata.state.frac_done * 100);
//  if(run_i==10){
		for(unsigned i=0;i<fft_len1;i++)
			fprintf(stderr,"data[%u][0]=\t%e\tdata[%u][1]=\t%e\n",i,data[i][0],i,data[i][1]);
		exit(0);
//  }
#endif
}

#endif

#if ( __SSE__ && SPLIT_COMPLEX)
static const inline void BlockPrefetch_4K(void* addr){
#if _GNU_SOURCE
asm("mov esi,[addr]");
asm("mov eax,16");//we do 64 bytes per mov that is 256 bytes per loop cycle 4*1024/256=16
asm("mov ecx,512");//512*8=4096 - total bytes, 32*8=256 - bytes per loop
asm("prefetchloop: ");//we do decrement to save cmp instruction and disable hardware prefetcher.
asm("mov edx,[esi+ecx*8-64]");
asm("mov edx,[esi+ecx*8-128+8]"); //shift less than 64 bytes to read from 2 different lines _and_ 2 different banks (1bank=8bytes)
asm("mov edx,[esi+ecx*8-192+16]");
asm("mov edx,[esi+ecx*8-256+24]");
asm("sub ecx,32");
asm("dec eax");
asm("jnz prefetchloop");
#else
#ifndef _WIN64
__asm{
	mov esi,[addr]
	;mov eax,16;we do 64 bytes per mov that is 256 bytes per loop cycle 4*1024/256=16
	mov ecx,512;512*8=4096 - total bytes, 32*8=256 - bytes per loop
prefetchloop: ;we do decrement to save cmp instruction and disable hardware prefetcher.
	mov edx,[esi+ecx*8-64]
	mov edx,[esi+ecx*8-128+8]; shift less than 64 bytes to read from 2 different lines _and_ 2 different banks (1bank=8bytes)
	mov edx,[esi+ecx*8-192+16]
	mov edx,[esi+ecx*8-256+24]
	sub ecx,32
	;dec eax
	jnz prefetchloop
}
#endif
#endif //_GNU_SOURCE
}
#endif // ( __SSE__ && SPLIT_COMPLEX)

/* Science::dechirp
 * Now DIVIDES by the chirp function, instead of multiplying.
 *
 * Arguments:
 * dm: dispersion measure to dechirp by, in units of 0.4 microseconds
 * power: array of length fft_len/2, to hold the power in each time bin
 *        I'm not sure why we only record the first fft_len/2 bins.
 * sign: +1 for a positive dm, -1 for a negative dm
 */
void Science::dechirp(int dm, float *power, int sign) {
#if (__SSE3__ && !SPLIT_COMPLEX)
 #if !(( _GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
	//Timings<T_dechirp> dechirp_timer;
 #endif
	const size_t& fft_len = client.ap_shmem->ap_gdata.state.fft_len;
	int n = dm*fft_len;
	size_t i;
	static float avg_pow = 2*fft_len*float(fft_len);
	// avg_pow contains N^2 for unnormalized fft, and 2 for 1-bit
	// convention
 #if BOINC_APP_GRAPHICS
	client.status = "Dechirping.";
 #endif
 #if 0
	fprintf(stderr,"Input data\n");
	for(unsigned i=0;i<fft_len1/2;i++)
	  fprintf(stderr,"data[%u][0]=%e\tdata[%u][1]=%e\n",i,data[i][0],i,data[i][1]);
 #endif
 #if USE_GPU
	GPUdechirp(gputemp,gpudata,&(gpuchirp[n]),sign,n,fft_len);
 #elif __SSE3__
	//int fftend = (fft_len*2) - ((fft_len*2) & 3);
	__m128 vsign = _mm_set1_ps(sign <0 ? 1.0 : -1.0);//R:adopt sign for 5.04 change
	float *dataP = data[0];
	float *chirpP = chirp[n];
  #if USE_FFTW // was _WIN32 && USE_VIRTUAL_ALLOC
	float *tempP = temp_in[0];
  #else
	float *tempP = temp[0];
  #endif
  #if TWINDECHIRP
	#define NUMPERPAGE 1024 // 4096/sizeof(float)
	static const __m128 NEG_S = {-0.0f, 0.0f, -0.0f, 0.0f};

	if (negredy != dm) {
	  unsigned int kk, tlbmsk = fft_len*2-1;
	  __m128 tmp1, tmp2, tmp3;
	  float* tinP = temp_in_neg[0];

      if (modev505)
	  {
		  for (kk=0; kk<fft_len*2; kk += NUMPERPAGE) {
			tlbt1 = dataP[(kk+NUMPERPAGE)&tlbmsk]; // TLB priming
			tlbt2 = chirpP[(kk+NUMPERPAGE)&tlbmsk]; // TLB priming
			// prefetch entire blocks, one 32 byte P3 cache line per loop
			for (i=kk+8; i<kk+NUMPERPAGE; i+=8) {
			_mm_prefetch((char*)&dataP[i], _MM_HINT_NTA);
			_mm_prefetch((char*)&chirpP[i], _MM_HINT_NTA);
			}
			// process 4 floats per loop
			for (i=kk; i<kk+NUMPERPAGE; i+=4) {
			  tmp1=_mm_load_ps(&chirpP[i]);            //  s,  c
			  tmp2=_mm_load_ps(&dataP[i]);             //  i,  r
			  tmp3=tmp1;
			  tmp1=_mm_movehdup_ps(tmp1);              //  s,  s
			  tmp3=_mm_moveldup_ps(tmp3);              //  c,  c
			  tmp1=_mm_xor_ps(tmp1, NEG_S);            //  s, -s
			  tmp3=_mm_mul_ps(tmp3, tmp2);             // ic, rc
			  tmp2=_mm_shuffle_ps(tmp2, tmp2, 0xb1);   //  r,  i
			  tmp1=_mm_mul_ps(tmp1, tmp2);             // rs,-is
			  tmp2=tmp1;
			  tmp2=_mm_add_ps(tmp2, tmp3);
			  _mm_store_ps(&tinP[i], tmp2);            // ic+rs, rc-is
			  tmp3=_mm_sub_ps(tmp3, tmp1);
			  _mm_store_ps(&tempP[i], tmp3);           // ic-rs, rc+is
			}
		  } //kk
	  }
	  else //v6 or newer swaps dm signs
	  {
		  for (kk=0; kk<fft_len*2; kk += NUMPERPAGE) {
			tlbt1 = dataP[(kk+NUMPERPAGE)&tlbmsk]; // TLB priming
			tlbt2 = chirpP[(kk+NUMPERPAGE)&tlbmsk]; // TLB priming
			// prefetch entire blocks, one 32 byte P3 cache line per loop
			for (i=kk+8; i<kk+NUMPERPAGE; i+=8) {
			_mm_prefetch((char*)&dataP[i], _MM_HINT_NTA);
			_mm_prefetch((char*)&chirpP[i], _MM_HINT_NTA);
			}
			// process 4 floats per loop
			for (i=kk; i<kk+NUMPERPAGE; i+=4) {
			  tmp1=_mm_load_ps(&chirpP[i]);            //  s,  c
			  tmp2=_mm_load_ps(&dataP[i]);             //  i,  r
			  tmp3=tmp1;
			  tmp1=_mm_movehdup_ps(tmp1);              //  s,  s
			  tmp3=_mm_moveldup_ps(tmp3);              //  c,  c
			  tmp1=_mm_xor_ps(tmp1, NEG_S);            //  s, -s
			  tmp3=_mm_mul_ps(tmp3, tmp2);             // ic, rc
			  tmp2=_mm_shuffle_ps(tmp2, tmp2, 0xb1);   //  r,  i
			  tmp1=_mm_mul_ps(tmp1, tmp2);             // rs,-is
			  tmp2=tmp1;
			  tmp2=_mm_sub_ps(tmp2, tmp3);
			  _mm_store_ps(&tinP[i], tmp2);            // ic-rs, rc+is
			  tmp3=_mm_add_ps(tmp3, tmp1);
			  _mm_store_ps(&tempP[i], tmp3);           // ic+rs, rc-is
			}
		  } //kk
	  }
	  negredy = dm;
	}

  #else
	//typedef union{
	//  __m128 v_f;
	//  __m128i v_i;
	//} v8;
	//Jason: Removed union casts, saved some regs, so unrolled once more to use them.
	// Rerolled after SSE build penalty for extra register usage in emulation showed itself.
	__m128 d1,c1,t1,t2;
//	__m128 d1a,c1a,t1a,t2a;
	for (i=0; i<fft_len*2; i+=/*8*/4) {
	  d1 = _mm_load_ps(dataP+i);
//	  d1a = _mm_load_ps(dataP+i+4);
	  c1 = _mm_load_ps(chirpP+i);
//	  c1a = _mm_load_ps(chirpP+i+4);
	  t1 = _mm_mul_ps(d1,_mm_moveldup_ps(c1));
//	  t1a = _mm_mul_ps(d1a,_mm_moveldup_ps(c1a));
	  t2 = _mm_mul_ps(_mm_shuffle_ps(d1,d1,0xb1), _mm_mul_ps(_mm_movehdup_ps(c1),vsign));
//	  t2a = _mm_mul_ps(_mm_shuffle_ps(d1a,d1a,0xb1), _mm_mul_ps(_mm_movehdup_ps(c1a),vsign));
	  _mm_store_ps(tempP+i,_mm_addsub_ps(t1,t2));
//	  _mm_store_ps(tempP+i+4,_mm_addsub_ps(t1a,t2a));
	}
  #endif

/*R: separate treating of end elements is not needed because fft_len is big power of 2 - no tail elements at all
	for ( i/=2; i<fft_len; i++) {

	  if (sign < 0) {//R: adopt sign for 5.04 changes
		temp[i][0] = data[i][0]*chirp[n+i][0] - data[i][1]*chirp[n+i][1];
		temp[i][1] = data[i][1]*chirp[n+i][0] + data[i][0]*chirp[n+i][1];
	  }

	  else {
		temp[i][0] = data[i][0]*chirp[n+i][0] + data[i][1]*chirp[n+i][1];
		temp[i][1] = data[i][1]*chirp[n+i][0] - data[i][0]*chirp[n+i][1];
	  }

	}
*/
/* #elif __SSE3__ && _GNU_SOURCE //uje: deactivating this part, as it gets replaced by TWINDECHIRP
	__m128 vsign = _mm_set1_ps(sign <0 ? 1.0 : -1.0);//R:adopt sign for 5.04 change
	float *dataP = data[0];
	float *chirpP = chirp[n];
	float *tempP = temp[0];

	//Jason: Removed union casts, saved some regs, so unrolled once more to use them.
	// Rerolled after SSE build penalty for extra register usage in emulation showed itself.
	__m128 d1,c1,t1,t2;
	for (i=0; i<fft_len*2; i+=4) { //i+=8
	  d1 = _mm_load_ps(dataP+i);
	  c1 = _mm_load_ps(chirpP+i);
	  t1 = _mm_mul_ps(d1,_mm_moveldup_ps(c1));
	  t2 = _mm_mul_ps(_mm_shuffle_ps(d1,d1,0xb1), _mm_mul_ps(_mm_movehdup_ps(c1),vsign));
	  _mm_store_ps(tempP+i,_mm_addsub_ps(t1,t2));
	} */
/*R: separate treating of end elements is not needed because fft_len is big power of 2 - no tail elements at all*/

 #else
	if(sign<0){//R: adopt sign for 5.04 changes
	  for (i=0; i<fft_len; i+=4) {
		  register size_t k=n+i;
		  temp[i][0] = data[i][0]*chirp[k][0] - data[i][1]*chirp[k][1];
		  temp[i][1] = data[i][1]*chirp[k][0] + data[i][0]*chirp[k][1];
		  temp[i+1][0] = data[i+1][0]*chirp[k+1][0] - data[i+1][1]*chirp[k+1][1];
		  temp[i+1][1] = data[i+1][1]*chirp[k+1][0] + data[i+1][0]*chirp[k+1][1];
		  temp[i+2][0] = data[i+2][0]*chirp[k+2][0] - data[i+2][1]*chirp[k+2][1];
		  temp[i+2][1] = data[i+2][1]*chirp[k+2][0] + data[i+2][0]*chirp[k+2][1];
		  temp[i+3][0] = data[i+3][0]*chirp[k+3][0] - data[i+3][1]*chirp[k+3][1];
		  temp[i+3][1] = data[i+3][1]*chirp[k+3][0] + data[i+3][0]*chirp[k+3][1];

	  }
	}else{
	  for (i=0; i<fft_len; i++) {
		temp[i][0] = data[i][0]*chirp[n+i][0] + data[i][1]*chirp[n+i][1];
		temp[i][1] = data[i][1]*chirp[n+i][0] - data[i][0]*chirp[n+i][1];
	  }
	}
 #endif
 #if BOINC_APP_GRAPHICS
	client.status = "Computing inverse FFT.";
 #endif
	{
 #if !(( _GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
//	Timings<T_FFT_inverse> timings;
 #endif
 #if 0
  #if USE_GPU
	gpuError=cudaMemcpy(temp,gputemp,sizeof(ap_complex)*client.ap_shmem->ap_gdata.state.fft_len,cudaMemcpyDeviceToHost);
		if (!gpuError==0)
		{
			fprintf(stderr,"GPU memory copy error (data down) (temp,gputemp)...");
			exit(0);
		}
  #endif
	fprintf(stderr,"Before inverse FFT\n");
	  for(unsigned i=0;i<fft_len1;i++)
		fprintf(stderr,"temp[%u][0]=%e\ttemp[%u][1]=%e\n",i,temp[i][0],i,temp[i][1]);
 #endif

 #if USE_GPU
	gpufftError=cufftExecC2C(cuFFTplan,gputemp,gputemp,CUFFT_INVERSE);
	if (!gpufftError==CUFFT_SUCCESS)
	  {
	  fprintf(stderr,"GPU FFT error ...");
	  exit(0);
	  }
 #elif USE_FFTW
  #if TWINDECHIRP
	if (sign >=0) fftwf_execute(ip);
	else {
	  fftwf_execute(ipn);
	  negredy = -1;
	}
  #else
	fftwf_execute(ip);
  #endif
 #elif USE_IPP
  ippsFFTInv_CToC_32fc_I((Ipp32fc*)temp,pFFTSpec,NULL); //R: maybe it's worth to provide some buffer for FFT...
 #elif USE_ACML
	cfft1d(1,acmlfftlen,(complex*)temp,ip,&info);
 #else
	#pragma error "FFT library needed"
#endif
  }//inverse FFT
  /* only save 1/2 of the fft ... need to check this */
#if 0
#if USE_GPU
	gpuError=cudaMemcpy(temp,gputemp,sizeof(ap_complex)*client.ap_shmem->ap_gdata.state.fft_len,cudaMemcpyDeviceToHost);
		if (!gpuError==0)
		{
			fprintf(stderr,"GPU memory copy error (data down) (temp,gputemp)...");
			exit(0);
		}
#endif
	fprintf(stderr,"After inverse FFT\n");
		for(unsigned i=0;i<fft_len1;i++)
			fprintf(stderr,"temp[%u][0]=%e\ttemp[%u][1]=%e\n",i,temp[i][0],i,temp[i][1]);

#endif

#if USE_GPU
	float ravg_pow=1.0/avg_pow;
	halfTemp(gputemp,gpupower,ravg_pow,fft_len);
   //gpuError=cudaMemcpy(power,gpupower,sizeof(float)*fft_len/2,cudaMemcpyDeviceToHost);
	//  if (!gpuError==cudaSuccess)
	//  {
	//      fprintf(stderr,"GPU memory copy error (data down) ...");
	//      exit(0);
	//  }
#elif __SSE3__ && !( _GNU_SOURCE || __APPLE__ || __FreeBSD__ )
  __m128 ravg_pow = _mm_set1_ps(1.0/avg_pow);
//R: fft_len big power of 2, bigger than 8 for sure. So, we don't need any checks for ending elements.
  //fftend = (fft_len/2) - ((fft_len/2) & 3);
  for (i=0; i<fft_len/2; i+=4) {
	  __m128 a1 = _mm_load_ps(temp[i]);
	  __m128 a2 = _mm_load_ps(temp[i+2]);
	  __m128 r1 = _mm_mul_ps(_mm_hadd_ps(_mm_mul_ps(a1,a1),_mm_mul_ps(a2,a2)), ravg_pow);
	  _mm_store_ps(&client.power[i], r1);
  }

  //for (; i<fft_len/2b; i++) {
//    client.power[i] = (float)((temp[i][0]*temp[i][0] + temp[i][1]*temp[i][1])/avg_pow);
 // }

#elif __SSE3__ && ( _GNU_SOURCE || __APPLE__ || __FreeBSD__ )
  __m128 ravg_pow = _mm_set1_ps(1.0/avg_pow);
//R: fft_len big power of 2, bigger than 8 for sure. So, we don't need any checks for ending elements.
  for (i=0; i<fft_len/2; i+=4) {
	  const float* tP1 = (float*)&temp[i]; //uje: const float* needed at compile time
	  const float* tP2 = (float*)&temp[i+2]; //uje: const float* needed
	  __m128 a1 = _mm_loadu_ps(tP1); // uje: unaligned !
	  __m128 a2 = _mm_loadu_ps(tP2); //uje: unaligned !
	  __m128 r1 = _mm_mul_ps(_mm_hadd_ps(_mm_mul_ps(a1,a1),_mm_mul_ps(a2,a2)), ravg_pow);
	  _mm_store_ps(&client.power[i], r1);
  }

#else
  for (i=0; i<fft_len/2; i++) {
	client.power[i] = (float)((temp[i][0]*temp[i][0] + temp[i][1]*temp[i][1])/avg_pow);
  }
#endif
#if 0
	static int run_i;
	fprintf(stderr,"%d dechirp run finished @%.2f percent.\n",run_i,client.ap_shmem->ap_gdata.state.frac_done * 100);
	if(run_i++==100000){//R:can set some another loop iteration instead 0
		for(unsigned i=0;i<fft_len1/2;i++)
			fprintf(stderr,"power[%u]=%e\n",i,power[i]);
		exit(0);
	}

#endif
#else  // !(__SSE3__ && !SPLIT_COMPLEX)
#if !(( _GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
  //Timings<T_dechirp> dechirp_timer;
#endif
  const size_t& fft_len = client.ap_shmem->ap_gdata.state.fft_len;
  int n = dm*fft_len;
  size_t i;
  static float avg_pow = 2*fft_len*float(fft_len);
  // avg_pow contains N^2 for unnormalized fft, and 2 for 1-bit
  // convention
#if BOINC_APP_GRAPHICS
  client.status = "Dechirping.";
#endif
  /* For debugging.  Actually I don't think this runs fast enough
  static int idum = -1;
  for(i=0; i< fft_len; i++) {
   data[i][0] = random_discrete_sum(fft_len, idum);
   data[i][1] = random_discrete_sum(fft_len, idum);
  } */
#if 0 //__SSE3__
  int fftend = (fft_len*2) - ((fft_len*2) & 3);
  __m128 vsign = _mm_set1_ps(sign <0 ? 1.0 : -1.0);//R: adopt sign for 5.04 changes
  float *dataP = data[0];
  float *chirpP = chirp[n];
  float *tempP = temp[0];
	typedef union{
		__m128 v_f;
		__m128i v_i;
	} v8;
  for (i=0; i<fftend; i+=4) {
	  v8 d1;
		d1.v_f= _mm_load_ps(dataP+i);
	  __m128 c1 = _mm_load_ps(chirpP+i);
	  __m128 t1 = _mm_mul_ps(d1.v_f,_mm_moveldup_ps(c1));
		v8 x;x.v_i=(_mm_shuffle_epi32(d1.v_i,0xb1));
	  __m128 t2 = _mm_mul_ps(x.v_f,_mm_mul_ps(_mm_movehdup_ps(c1),vsign));
//    __m128 t2 = _mm_mul_ps(_mm_shuffle_ps(d1,d1,0xb1),_mm_mul_ps(_mm_movehdup_ps(c1),vsign));
	  _mm_store_ps(tempP+i,_mm_addsub_ps(t1,t2));
  }
  for ( i/=2; i<fft_len; i++) {
	/* positive DM */
	if (sign < 0) {//R: adopt sign for 5.04 changes
	  temp[i][0] = data[i][0]*chirp[n+i][0] - data[i][1]*chirp[n+i][1];
	  temp[i][1] = data[i][1]*chirp[n+i][0] + data[i][0]*chirp[n+i][1];
	}
	/* negative DM */
	else {
	  temp[i][0] = data[i][0]*chirp[n+i][0] + data[i][1]*chirp[n+i][1];
	  temp[i][1] = data[i][1]*chirp[n+i][0] - data[i][0]*chirp[n+i][1];
	}
  }
#else
#if 0 //USE_GPU
	  GPUdechirp(gputemp,gpudata,gpuchirp,sign,n,fft_len);
#else
  #if _WIN32 && USE_VIRTUAL_ALLOC
	 ap_complex *tempP = temp_in;
  #else
	 ap_complex *tempP = temp;
  #endif
  if(sign<0){//R: adopt sign for 5.04 changes
#if ( __SSE__ && SPLIT_COMPLEX) //uje:  && !_GNU_SOURCE needs repair
		register float* dpr=data;
		register float* dpi=i_data;
		register float* cpr=chirp+n;
		register float* cpi=i_chirp+n;
		register float* tpr=temp;
		register float* tpi=i_temp;
		for(;dpr<data+fft_len;dpr+=1024,dpi+=1024,
			cpr+=1024,cpi+=1024,
			tpr+=1024,tpi+=1024){

				BlockPrefetch_4K(dpr);
				BlockPrefetch_4K(dpi);
				BlockPrefetch_4K(cpr);
				BlockPrefetch_4K(cpi);
				BlockPrefetch_4K(tpr);
				BlockPrefetch_4K(tpi);
			  for (i=0; i<1024; i+=4) {

			__m128 dr=_mm_load_ps(dpr+i);//real part of data[i]
			__m128 di=_mm_load_ps(dpi+i);//imaginary -"-
			__m128 cr=_mm_load_ps(cpr+i);
			__m128 ci=_mm_load_ps(cpi+i);


		   _mm_store_ps(tpr+i, _mm_sub_ps(_mm_mul_ps(dr,cr), _mm_mul_ps(di,ci)));
		   _mm_store_ps(tpi+i, _mm_add_ps(_mm_mul_ps(di,cr), _mm_mul_ps(dr,ci)));
			  }
		}

#else
		for (i=0; i<fft_len; i+=4) {
			register size_t k=n+i;
			tempP[i][0] = data[i][0]*chirp[k][0] - data[i][1]*chirp[k][1];
			tempP[i][1] = data[i][1]*chirp[k][0] + data[i][0]*chirp[k][1];
			tempP[i+1][0] = data[i+1][0]*chirp[k+1][0] - data[i+1][1]*chirp[k+1][1];
			tempP[i+1][1] = data[i+1][1]*chirp[k+1][0] + data[i+1][0]*chirp[k+1][1];
			tempP[i+2][0] = data[i+2][0]*chirp[k+2][0] - data[i+2][1]*chirp[k+2][1];
			tempP[i+2][1] = data[i+2][1]*chirp[k+2][0] + data[i+2][0]*chirp[k+2][1];
			tempP[i+3][0] = data[i+3][0]*chirp[k+3][0] - data[i+3][1]*chirp[k+3][1];
			tempP[i+3][1] = data[i+3][1]*chirp[k+3][0] + data[i+3][0]*chirp[k+3][1];
	  }
#endif
  }else{
#if ( __SSE__ && SPLIT_COMPLEX) //uje: && !_GNU_SOURCE needs repair
		register float* dpr=data;
		register float* dpi=i_data;
		register float* cpr=chirp+n;
		register float* cpi=i_chirp+n;
		register float* tpr=temp;
		register float* tpi=i_temp;
		for(;dpr<data+fft_len;dpr+=1024,dpi+=1024,
			cpr+=1024,cpi+=1024,
			tpr+=1024,tpi+=1024){

				BlockPrefetch_4K(dpr);
				BlockPrefetch_4K(dpi);
				BlockPrefetch_4K(cpr);
				BlockPrefetch_4K(cpi);
				BlockPrefetch_4K(tpr);
				BlockPrefetch_4K(tpi);

			for (i=0; i<1024; i+=4) {
				__m128 dr=_mm_load_ps(dpr+i);//real part of data[i]
				__m128 di=_mm_load_ps(dpi+i);//imaginary -"-
				__m128 cr=_mm_load_ps(cpr+i);
				__m128 ci=_mm_load_ps(cpi+i);
				_mm_store_ps(tpr+i,
				_mm_add_ps(_mm_mul_ps(dr,cr),
				_mm_mul_ps(di,ci)));
				_mm_store_ps(tpr+i, _mm_add_ps(_mm_mul_ps(dr,cr), _mm_mul_ps(di,ci)));
				_mm_store_ps(tpi+i, _mm_sub_ps(_mm_mul_ps(di,cr), _mm_mul_ps(dr,ci)));
			}
		}

#else
	for (i=0; i<fft_len; i+=4) {
		register size_t k=n+i;
		tempP[i][0] = data[i][0]*chirp[k][0] + data[i][1]*chirp[k][1];
		tempP[i][1] = data[i][1]*chirp[k][0] - data[i][0]*chirp[k][1];
		tempP[i+1][0] = data[i+1][0]*chirp[k+1][0] + data[i+1][1]*chirp[k+1][1];
		tempP[i+1][1] = data[i+1][1]*chirp[k+1][0] - data[i+1][0]*chirp[k+1][1];
		tempP[i+2][0] = data[i+2][0]*chirp[k+2][0] + data[i+2][1]*chirp[k+2][1];
		tempP[i+2][1] = data[i+2][1]*chirp[k+2][0] - data[i+2][0]*chirp[k+2][1];
		tempP[i+3][0] = data[i+3][0]*chirp[k+3][0] + data[i+3][1]*chirp[k+3][1];
		tempP[i+3][1] = data[i+3][1]*chirp[k+3][0] - data[i+3][0]*chirp[k+3][1];
	}
#endif

  }
#endif
#endif//non SSE3
  {
#if !(( _GNU_SOURCE && (__WORDSIZE == 32)) || __APPLE__ )
//Timings<T_FFT_inverse> timings;
#endif
#if BOINC_APP_GRAPHICS
  client.status = "Computing inverse FFT.";
#endif
#if 0 //USE_GPU
  gpuError=cudaMemcpy(gpudata,temp,sizeof(ap_complex)*fft_len,cudaMemcpyHostToDevice);
	if (!gpuError==cudaSuccess)
		{
			fprintf(stderr,"GPU memory copy error (data up) ...");
			exit(0);
		}

  gpufftError=cufftExecC2C(cuFFTplan,gputemp,gputemp,CUFFT_INVERSE);
		if (!gpufftError==CUFFT_SUCCESS)
		{
			fprintf(stderr,"GPU FFT error ...");
			exit(0);
		}

  gpuError=cudaMemcpy(temp,gputemp,sizeof(ap_complex)*fft_len,cudaMemcpyDeviceToHost);
		if (!gpuError==cudaSuccess)
		{
			fprintf(stderr,"GPU memory copy error (data down) ...");
			exit(0);
		}

#elif USE_FFTW
	fftwf_execute(ip);
#elif USE_IPP
  ippsFFTInv_CToC_32fc_I((Ipp32fc*)temp,pFFTSpec,NULL); //R: maybe it's worth to provide some buffer for FFT...
#else
   #pragma error "FFT library needed"
#endif
  }
  /* only save 1/2 of the fft ... need to check this */
#if ( __SSE__ && SPLIT_COMPLEX) //uje:  && !_GNU_SOURCE
	for(i=0; i<fft_len/2; i+=4) {

		__m128 tr=_mm_load_ps(temp+i);
		__m128 ti=_mm_load_ps(i_temp+i);
		__m128 ravg=_mm_set1_ps(1.0/avg_pow);
/*#elif ( __SSE__ && SPLIT_COMPLEX) //uje:  && _GNU_SOURCE for later making use of SPLIT_COMPLEX work
	for(i=0; i<fft_len/2; i+=4) {

		__m128 tr=_mm_load_ps(temp+i);
		__m128 ti=_mm_load_ps(i_temp+i);
		__m128 ravg=_mm_set1_ps(1.0/avg_pow);   */
#ifndef BOINC_APP_GRAPHICS
		//R:use same pointer for both temp and client.power arrays with non-graphic build.
		_mm_store_ps(temp+i,_mm_mul_ps(_mm_add_ps(_mm_mul_ps(tr,tr), _mm_mul_ps(ti,ti)),ravg));
#else
		_mm_store_ps(client.power+i, _mm_mul_ps(_mm_add_ps(_mm_mul_ps(tr,tr), _mm_mul_ps(ti,ti)),ravg));
#endif
	}/*
	for (i=0; i<fft_len/2; i++) {
		client.power[i] = (float)((temp[i]*temp[i] + i_temp[i]*i_temp[i])/avg_pow);
	}*/

#if 0 //USE_GPU
	halfTemp(gputemp,gpupower,avg_pow,fft_len/2);
	gpuError=cudaMemcpy(power,gpupower,sizeof(ap_complex)*fft_len/2,cudaMemcpyDeviceToHost);
	if (!gpuError==cudaSuccess) {
			fprintf(stderr,"GPU memory copy error (data down) ...");
			exit(0);
	}
#else
	for (i=0; i<fft_len/2; i++) {
		client.power[i] = (float)((temp[i][0]*temp[i][0] + temp[i][1]*temp[i][1])/avg_pow);
	}
#endif
#endif
#endif // all variants
#if 0
	{
	float power_sum=0;
	//R: Some debug calculation of actual average power.

	for(unsigned i=0;i<fft_len/2;i++) power_sum+=power[i];
	Counters<T_AveragePower,float>::update(power_sum*2/fft_len);
	}
#endif
}

void* malloc_a(size_t size){
#if USE_FFTW
	return fftwf_malloc(size);
#elif USE_IPP
	return ippMalloc(size);
#elif USE_ACML
	return fwMalloc(size);
#elif USE_GPU
	void* temp;
	cudaMallocHost((void**)&temp,size);
#ifdef __SSE__
	if(temp!=(void*)((unsigned)temp&(~0xF))){
		fprintf(stderr,"ERROR: Unaligned memory allocation, %p;%u\n",temp,temp);
		exit(0);
	}

#endif
	return temp;

#else
	#error Need aligned malloc
#endif
}

void free_a(void* ptr){
#if USE_FFTW
	fftwf_free(ptr);
#elif USE_IPP
	ippFree(ptr);
#elif USE_GPU
	cudaFree(ptr);
#elif USE_ACML
	return fwFree(ptr);
#else
#error Need aligned free
#endif
}

/* coadd_in_place
 *
 * Arguments:
 * power: an array of power values, obtained after dechirping
 * size: the number of floats in the initial power array.
 * (After the coadd, the number of floats will be halved.)
 *
 * This function halves the size of the power array by adding
 * array elements in consecutive pairs.
 */
void Science::coadd_in_place(float *power, int size) {
	//Timings<T_coadd> coadd_timing;

	//static timer coadd_timer;
	//coadd_timer.start();
	//fprintf(r_debug,"Before coadd\n");
	//for(int i=0;i<size;i++){
	//  fprintf(r_debug,"power[%d]=%f\n",i,power[i]);
	//}
#if USE_GPU
	GPU_coadd_in_place(gpupower,size);
#elif __SSE3__ && !( _GNU_SOURCE || __APPLE__ || __FreeBSD__ )
  for(int i=0;i<size/2;i+=8){
	  _mm_store_ps(&power[i],_mm_hadd_ps(_mm_load_ps(&power[2*i]), _mm_load_ps(&power[2*i+4])));
	  _mm_store_ps(&power[i+4],_mm_hadd_ps(_mm_load_ps(&power[2*i+8]), _mm_load_ps(&power[2*i+12])));
  }
#elif __SSE3__ && ( _GNU_SOURCE || __APPLE__ || __FreeBSD__ )
  for(int i=0;i<size/2;i+=16){
	  _mm_store_ps(&power[i], _mm_hadd_ps(_mm_load_ps(&power[2*i]), _mm_load_ps(&power[2*i+4])));
	  _mm_store_ps(&power[i+4], _mm_hadd_ps(_mm_load_ps(&power[2*i+8]), _mm_load_ps(&power[2*i+12])));
    _mm_store_ps(&power[i+8], _mm_hadd_ps(_mm_load_ps(&power[2*i+16]), _mm_load_ps(&power[2*i+20])));
    _mm_store_ps(&power[i+12], _mm_hadd_ps(_mm_load_ps(&power[2*i+24]), _mm_load_ps(&power[2*i+28])));
  }
#else
  int i;
  for (i=0; i<size/2; i++) {
	power[i] = power[2*i] + power[2*i+1];
  }
#endif
  //coadd_timer.stop();
  //fprintf(r_debug,"coadd_in_place=%ld ticks, total=%ld\n",coadd_timer.get_ticks(),coadd_timer.get_total_ticks());
	//fprintf(r_debug,"After coadd\n");
	//for(int i=0;i<size;i++){
	//  fprintf(r_debug,"power[%d]=%f\n",i,power[i]);
	//}

}

#if _SSE_ || (__SSE__ && ( _GNU_SOURCE || __APPLE__ || __FreeBSD__ ))
#ifdef _MSC_VER
# define ALIGN16_BEG __declspec(align(16))
# define ALIGN16_END
#else /* gcc or icc */
# define ALIGN16_BEG
# define ALIGN16_END __attribute__((aligned(16)))
#endif

typedef __m128 v4sf;
typedef __m64 v2si;
#define _PS_CONST(Name, Val)                                            \
  static const ALIGN16_BEG float _ps_##Name[4] ALIGN16_END = { Val, Val, Val, Val }
#define _PI32_CONST(Name, Val)                                            \
  static const ALIGN16_BEG int _pi32_##Name[4] ALIGN16_END = { Val, Val, Val, Val }
#define _PS_CONST_TYPE(Name, Type, Val)                                 \
  static const ALIGN16_BEG Type _ps_##Name[4] ALIGN16_END = { Val, Val, Val, Val }

_PS_CONST(1  , 1.0f);
_PS_CONST(0p5, 0.5f);
_PS_CONST_TYPE(min_norm_pos, int, 0x00800000);
_PS_CONST_TYPE(mant_mask, int, 0x7f800000);
_PS_CONST_TYPE(inv_mant_mask, int, ~0x7f800000);
_PS_CONST_TYPE(sign_mask, int, 0x80000000);
_PS_CONST_TYPE(inv_sign_mask, int, ~0x80000000);
_PI32_CONST(1, 1);
_PI32_CONST(inv1, ~1);
_PI32_CONST(2, 2);
_PI32_CONST(4, 4);
_PI32_CONST(0x7f, 0x7f);
_PS_CONST(cephes_SQRTHF, 0.707106781186547524);
_PS_CONST(cephes_log_p0, 7.0376836292E-2);
_PS_CONST(cephes_log_p1, - 1.1514610310E-1);
_PS_CONST(cephes_log_p2, 1.1676998740E-1);
_PS_CONST(cephes_log_p3, - 1.2420140846E-1);
_PS_CONST(cephes_log_p4, + 1.4249322787E-1);
_PS_CONST(cephes_log_p5, - 1.6668057665E-1);
_PS_CONST(cephes_log_p6, + 2.0000714765E-1);
_PS_CONST(cephes_log_p7, - 2.4999993993E-1);
_PS_CONST(cephes_log_p8, + 3.3333331174E-1);
_PS_CONST(cephes_log_q1, -2.12194440e-4);
_PS_CONST(cephes_log_q2, 0.693359375);

#if defined (__MINGW32__)

/* many versions of gcc used to be completely buggy with respect to some intrinsics
   The movehl_ps is fixed in mingw 3.4.5, but I found out that all the _mm_cmp* intrinsics were completely
   broken on my mingw gcc 3.4.5 ...

   Note that the bug on _mm_cmp* does occur only at -O0 optimization level
*/

inline __m128 my_movehl_ps(__m128 a, const __m128 b) {
	asm (
			"movhlps %2,%0\n\t"
			: "=x" (a)
			: "0" (a), "x"(b)
		);
	return a;                                 }
#warning "redefined _mm_movehl_ps (see gcc bug 21179)"
#define _mm_movehl_ps my_movehl_ps

inline __m128 my_cmplt_ps(__m128 a, const __m128 b) {
	asm (
			"cmpltps %2,%0\n\t"
			: "=x" (a)
			: "0" (a), "x"(b)
		);
	return a;
				  }
inline __m128 my_cmpgt_ps(__m128 a, const __m128 b) {
	asm (
			"cmpnleps %2,%0\n\t"
			: "=x" (a)
			: "0" (a), "x"(b)
		);
	return a;
}
inline __m128 my_cmpeq_ps(__m128 a, const __m128 b) {
	asm (
			"cmpeqps %2,%0\n\t"
			: "=x" (a)
			: "0" (a), "x"(b)
		);
	return a;
}
#warning "redefined _mm_cmpxx_ps functions..."
#define _mm_cmplt_ps my_cmplt_ps
#define _mm_cmpgt_ps my_cmpgt_ps
#define _mm_cmpeq_ps my_cmpeq_ps
#endif

typedef union xmm_mm_union {
  __m128 xmm;
  __m64 mm[2];
} xmm_mm_union;

#define COPY_XMM_TO_MM(xmm_, mm0_, mm1_) {          \
	xmm_mm_union u; u.xmm = xmm_;                   \
	mm0_ = u.mm[0];                                 \
	mm1_ = u.mm[1];                                 \
}

#define COPY_MM_TO_XMM(mm0_, mm1_, xmm_) {                         \
	xmm_mm_union u; u.mm[0]=mm0_; u.mm[1]=mm1_; xmm_ = u.xmm;      \
  }


static inline v4sf log_ps(v4sf x) {
  v2si mm0, mm1;
  v4sf one = *(v4sf*)_ps_1;
  v4sf invalid_mask = _mm_cmple_ps(x, _mm_setzero_ps());
  x = _mm_max_ps(x, *(v4sf*)_ps_min_norm_pos);  /* cut off denormalized stuff */
  COPY_XMM_TO_MM(x, mm0, mm1);
  mm0 = _mm_srli_pi32(mm0, 23);
  mm1 = _mm_srli_pi32(mm1, 23);
  x = _mm_and_ps(x, *(v4sf*)_ps_inv_mant_mask);
  x = _mm_or_ps(x, *(v4sf*)_ps_0p5);
  mm0 = _mm_sub_pi32(mm0, *(v2si*)_pi32_0x7f);
  mm1 = _mm_sub_pi32(mm1, *(v2si*)_pi32_0x7f);
  v4sf e = _mm_cvtpi32x2_ps(mm0, mm1);
  e = _mm_add_ps(e, one);
  v4sf mask = _mm_cmplt_ps(x, *(v4sf*)_ps_cephes_SQRTHF);
  v4sf tmp = _mm_and_ps(x, mask);
  x = _mm_sub_ps(x, one);
  e = _mm_sub_ps(e, _mm_and_ps(one, mask));
  x = _mm_add_ps(x, tmp);
  v4sf z = _mm_mul_ps(x,x);
  v4sf y = *(v4sf*)_ps_cephes_log_p0;
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_log_p1);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_log_p2);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_log_p3);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_log_p4);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_log_p5);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_log_p6);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_log_p7);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_log_p8);
  y = _mm_mul_ps(y, x);
  y = _mm_mul_ps(y, z);
  tmp = _mm_mul_ps(e, *(v4sf*)_ps_cephes_log_q1);
  y = _mm_add_ps(y, tmp);
  tmp = _mm_mul_ps(z, *(v4sf*)_ps_0p5);
  y = _mm_sub_ps(y, tmp);
  tmp = _mm_mul_ps(e, *(v4sf*)_ps_cephes_log_q2);
  x = _mm_add_ps(x, y);
  x = _mm_add_ps(x, tmp);
  x = _mm_or_ps(x, invalid_mask);
  _mm_empty();
  return x;
}

_PS_CONST(exp_hi,   88.3762626647949f);
_PS_CONST(exp_lo,   -88.3762626647949f);
_PS_CONST(cephes_LOG2EF, 1.44269504088896341);
_PS_CONST(cephes_exp_C1, 0.693359375);
_PS_CONST(cephes_exp_C2, -2.12194440e-4);
_PS_CONST(cephes_exp_p0, 1.9875691500E-4);
_PS_CONST(cephes_exp_p1, 1.3981999507E-3);
_PS_CONST(cephes_exp_p2, 8.3334519073E-3);
_PS_CONST(cephes_exp_p3, 4.1665795894E-2);
_PS_CONST(cephes_exp_p4, 1.6666665459E-1);
_PS_CONST(cephes_exp_p5, 5.0000001201E-1);

static inline v4sf exp_ps(v4sf x) {
  v4sf tmp = _mm_setzero_ps(), fx;
  v2si mm0, mm1;
  v4sf one = *(v4sf*)_ps_1;
  x = _mm_min_ps(x, *(v4sf*)_ps_exp_hi);
  x = _mm_max_ps(x, *(v4sf*)_ps_exp_lo);
  fx = _mm_mul_ps(x, *(v4sf*)_ps_cephes_LOG2EF);
  fx = _mm_add_ps(fx, *(v4sf*)_ps_0p5);
  tmp = _mm_movehl_ps(tmp, fx);
  mm0 = _mm_cvttps_pi32(fx);
  mm1 = _mm_cvttps_pi32(tmp);
  tmp = _mm_cvtpi32x2_ps(mm0, mm1);
  v4sf mask = _mm_cmpgt_ps(tmp, fx);
  mask = _mm_and_ps(mask, one);
  fx = _mm_sub_ps(tmp, mask);
  tmp = _mm_mul_ps(fx, *(v4sf*)_ps_cephes_exp_C1);
  v4sf z = _mm_mul_ps(fx, *(v4sf*)_ps_cephes_exp_C2);
  x = _mm_sub_ps(x, tmp);
  x = _mm_sub_ps(x, z);
  z = _mm_mul_ps(x,x);
  v4sf y = *(v4sf*)_ps_cephes_exp_p0;
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_exp_p1);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_exp_p2);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_exp_p3);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_exp_p4);
  y = _mm_mul_ps(y, x);
  y = _mm_add_ps(y, *(v4sf*)_ps_cephes_exp_p5);
  y = _mm_mul_ps(y, z);
  y = _mm_add_ps(y, x);
  y = _mm_add_ps(y, one);
  z = _mm_movehl_ps(z, fx);
  mm0 = _mm_cvttps_pi32(fx);
  mm1 = _mm_cvttps_pi32(z);
  mm0 = _mm_add_pi32(mm0, *(v2si*)_pi32_0x7f);
  mm1 = _mm_add_pi32(mm1, *(v2si*)_pi32_0x7f);
  mm0 = _mm_slli_pi32(mm0, 23);
  mm1 = _mm_slli_pi32(mm1, 23);

  v4sf pow2n;
  COPY_MM_TO_XMM(mm0, mm1, pow2n);

  y = _mm_mul_ps(y, pow2n);
  _mm_empty();
  return y;
}

_PS_CONST(minus_cephes_DP1, -0.78515625);
_PS_CONST(minus_cephes_DP2, -2.4187564849853515625e-4);
_PS_CONST(minus_cephes_DP3, -3.77489497744594108e-8);
_PS_CONST(sincof_p0, -1.9515295891E-4);
_PS_CONST(sincof_p1,  8.3321608736E-3);
_PS_CONST(sincof_p2, -1.6666654611E-1);
_PS_CONST(coscof_p0,  2.443315711809948E-005);
_PS_CONST(coscof_p1, -1.388731625493765E-003);
_PS_CONST(coscof_p2,  4.166664568298827E-002);
_PS_CONST(cephes_FOPI, 1.27323954473516);


static inline v4sf sin_ps(v4sf x) {
  v4sf xmm1, xmm2 = _mm_setzero_ps(), xmm3, sign_bit, y;
  v2si mm0, mm1, mm2, mm3;
  sign_bit = x;
  x = _mm_and_ps(x, *(v4sf*)_ps_inv_sign_mask);
  sign_bit = _mm_and_ps(sign_bit, *(v4sf*)_ps_sign_mask);
  y = _mm_mul_ps(x, *(v4sf*)_ps_cephes_FOPI);
  xmm2 = _mm_movehl_ps(xmm2, y);
  mm2 = _mm_cvttps_pi32(y);
  mm3 = _mm_cvttps_pi32(xmm2);
  mm2 = _mm_add_pi32(mm2, *(v2si*)_pi32_1);
  mm3 = _mm_add_pi32(mm3, *(v2si*)_pi32_1);
  mm2 = _mm_and_si64(mm2, *(v2si*)_pi32_inv1);
  mm3 = _mm_and_si64(mm3, *(v2si*)_pi32_inv1);
  y = _mm_cvtpi32x2_ps(mm2, mm3);
  mm0 = _mm_and_si64(mm2, *(v2si*)_pi32_4);
  mm1 = _mm_and_si64(mm3, *(v2si*)_pi32_4);
  mm0 = _mm_slli_pi32(mm0, 29);
  mm1 = _mm_slli_pi32(mm1, 29);
  mm2 = _mm_and_si64(mm2, *(v2si*)_pi32_2);
  mm3 = _mm_and_si64(mm3, *(v2si*)_pi32_2);
  mm2 = _mm_cmpeq_pi32(mm2, _mm_setzero_si64());
  mm3 = _mm_cmpeq_pi32(mm3, _mm_setzero_si64());
  v4sf swap_sign_bit, poly_mask;
  COPY_MM_TO_XMM(mm0, mm1, swap_sign_bit);
  COPY_MM_TO_XMM(mm2, mm3, poly_mask);
  sign_bit = _mm_xor_ps(sign_bit, swap_sign_bit);
  xmm1 = *(v4sf*)_ps_minus_cephes_DP1;
  xmm2 = *(v4sf*)_ps_minus_cephes_DP2;
  xmm3 = *(v4sf*)_ps_minus_cephes_DP3;
  xmm1 = _mm_mul_ps(y, xmm1);
  xmm2 = _mm_mul_ps(y, xmm2);
  xmm3 = _mm_mul_ps(y, xmm3);
  x = _mm_add_ps(x, xmm1);
  x = _mm_add_ps(x, xmm2);
  x = _mm_add_ps(x, xmm3);
  y = *(v4sf*)_ps_coscof_p0;
  v4sf z = _mm_mul_ps(x,x);
  y = _mm_mul_ps(y, z);
  y = _mm_add_ps(y, *(v4sf*)_ps_coscof_p1);
  y = _mm_mul_ps(y, z);
  y = _mm_add_ps(y, *(v4sf*)_ps_coscof_p2);
  y = _mm_mul_ps(y, z);
  y = _mm_mul_ps(y, z);
  v4sf tmp = _mm_mul_ps(z, *(v4sf*)_ps_0p5);
  y = _mm_sub_ps(y, tmp);
  y = _mm_add_ps(y, *(v4sf*)_ps_1);
  v4sf y2 = *(v4sf*)_ps_sincof_p0;
  y2 = _mm_mul_ps(y2, z);
  y2 = _mm_add_ps(y2, *(v4sf*)_ps_sincof_p1);
  y2 = _mm_mul_ps(y2, z);
  y2 = _mm_add_ps(y2, *(v4sf*)_ps_sincof_p2);
  y2 = _mm_mul_ps(y2, z);
  y2 = _mm_mul_ps(y2, x);
  y2 = _mm_add_ps(y2, x);
  xmm3 = poly_mask;
  y2 = _mm_and_ps(xmm3, y2);
  y = _mm_andnot_ps(xmm3, y);
  y = _mm_add_ps(y,y2);
  y = _mm_xor_ps(y, sign_bit);
  _mm_empty();
  return y;
}

static inline v4sf cos_ps(v4sf x) {
  v4sf xmm1, xmm2 = _mm_setzero_ps(), xmm3, y;
  v2si mm0, mm1, mm2, mm3;
  x = _mm_and_ps(x, *(v4sf*)_ps_inv_sign_mask);
  y = _mm_mul_ps(x, *(v4sf*)_ps_cephes_FOPI);
  xmm2 = _mm_movehl_ps(xmm2, y);
  mm2 = _mm_cvttps_pi32(y);
  mm3 = _mm_cvttps_pi32(xmm2);
  mm2 = _mm_add_pi32(mm2, *(v2si*)_pi32_1);
  mm3 = _mm_add_pi32(mm3, *(v2si*)_pi32_1);
  mm2 = _mm_and_si64(mm2, *(v2si*)_pi32_inv1);
  mm3 = _mm_and_si64(mm3, *(v2si*)_pi32_inv1);
  y = _mm_cvtpi32x2_ps(mm2, mm3);
  mm2 = _mm_sub_pi32(mm2, *(v2si*)_pi32_2);
  mm3 = _mm_sub_pi32(mm3, *(v2si*)_pi32_2);
  mm0 = _mm_andnot_si64(mm2, *(v2si*)_pi32_4);
  mm1 = _mm_andnot_si64(mm3, *(v2si*)_pi32_4);
  mm0 = _mm_slli_pi32(mm0, 29);
  mm1 = _mm_slli_pi32(mm1, 29);
  mm2 = _mm_and_si64(mm2, *(v2si*)_pi32_2);
  mm3 = _mm_and_si64(mm3, *(v2si*)_pi32_2);
  mm2 = _mm_cmpeq_pi32(mm2, _mm_setzero_si64());
  mm3 = _mm_cmpeq_pi32(mm3, _mm_setzero_si64());
  v4sf sign_bit, poly_mask;
  COPY_MM_TO_XMM(mm0, mm1, sign_bit);
  COPY_MM_TO_XMM(mm2, mm3, poly_mask);
  xmm1 = *(v4sf*)_ps_minus_cephes_DP1;
  xmm2 = *(v4sf*)_ps_minus_cephes_DP2;
  xmm3 = *(v4sf*)_ps_minus_cephes_DP3;
  xmm1 = _mm_mul_ps(y, xmm1);
  xmm2 = _mm_mul_ps(y, xmm2);
  xmm3 = _mm_mul_ps(y, xmm3);
  x = _mm_add_ps(x, xmm1);
  x = _mm_add_ps(x, xmm2);
  x = _mm_add_ps(x, xmm3);
  y = *(v4sf*)_ps_coscof_p0;
  v4sf z = _mm_mul_ps(x,x);
  y = _mm_mul_ps(y, z);
  y = _mm_add_ps(y, *(v4sf*)_ps_coscof_p1);
  y = _mm_mul_ps(y, z);
  y = _mm_add_ps(y, *(v4sf*)_ps_coscof_p2);
  y = _mm_mul_ps(y, z);
  y = _mm_mul_ps(y, z);
  v4sf tmp = _mm_mul_ps(z, *(v4sf*)_ps_0p5);
  y = _mm_sub_ps(y, tmp);
  y = _mm_add_ps(y, *(v4sf*)_ps_1);
  v4sf y2 = *(v4sf*)_ps_sincof_p0;
  y2 = _mm_mul_ps(y2, z);
  y2 = _mm_add_ps(y2, *(v4sf*)_ps_sincof_p1);
  y2 = _mm_mul_ps(y2, z);
  y2 = _mm_add_ps(y2, *(v4sf*)_ps_sincof_p2);
  y2 = _mm_mul_ps(y2, z);
  y2 = _mm_mul_ps(y2, x);
  y2 = _mm_add_ps(y2, x);
  xmm3 = poly_mask;
  y2 = _mm_and_ps(xmm3, y2);
  y = _mm_andnot_ps(xmm3, y);
  y = _mm_add_ps(y,y2);
  y = _mm_xor_ps(y, sign_bit);
  _mm_empty();
  return y;
}

static inline void sincos_ps(v4sf x, v4sf *s, v4sf *c) {
  v4sf xmm1, xmm2, xmm3 = _mm_setzero_ps(), sign_bit_sin, y;
  v2si mm0, mm1, mm2, mm3, mm4, mm5;
  sign_bit_sin = x;
  x = _mm_and_ps(x, *(v4sf*)_ps_inv_sign_mask);
  sign_bit_sin = _mm_and_ps(sign_bit_sin, *(v4sf*)_ps_sign_mask);
  y = _mm_mul_ps(x, *(v4sf*)_ps_cephes_FOPI);
  xmm3 = _mm_movehl_ps(xmm3, y);
  mm2 = _mm_cvttps_pi32(y);
  mm3 = _mm_cvttps_pi32(xmm3);
  mm2 = _mm_add_pi32(mm2, *(v2si*)_pi32_1);
  mm3 = _mm_add_pi32(mm3, *(v2si*)_pi32_1);
  mm2 = _mm_and_si64(mm2, *(v2si*)_pi32_inv1);
  mm3 = _mm_and_si64(mm3, *(v2si*)_pi32_inv1);
  y = _mm_cvtpi32x2_ps(mm2, mm3);
  mm4 = mm2;
  mm5 = mm3;
  mm0 = _mm_and_si64(mm2, *(v2si*)_pi32_4);
  mm1 = _mm_and_si64(mm3, *(v2si*)_pi32_4);
  mm0 = _mm_slli_pi32(mm0, 29);
  mm1 = _mm_slli_pi32(mm1, 29);
  v4sf swap_sign_bit_sin;
  COPY_MM_TO_XMM(mm0, mm1, swap_sign_bit_sin);
  mm2 = _mm_and_si64(mm2, *(v2si*)_pi32_2);
  mm3 = _mm_and_si64(mm3, *(v2si*)_pi32_2);
  mm2 = _mm_cmpeq_pi32(mm2, _mm_setzero_si64());
  mm3 = _mm_cmpeq_pi32(mm3, _mm_setzero_si64());
  v4sf poly_mask;
  COPY_MM_TO_XMM(mm2, mm3, poly_mask);
  xmm1 = *(v4sf*)_ps_minus_cephes_DP1;
  xmm2 = *(v4sf*)_ps_minus_cephes_DP2;
  xmm3 = *(v4sf*)_ps_minus_cephes_DP3;
  xmm1 = _mm_mul_ps(y, xmm1);
  xmm2 = _mm_mul_ps(y, xmm2);
  xmm3 = _mm_mul_ps(y, xmm3);
  x = _mm_add_ps(x, xmm1);
  x = _mm_add_ps(x, xmm2);
  x = _mm_add_ps(x, xmm3);
  mm4 = _mm_sub_pi32(mm4, *(v2si*)_pi32_2);
  mm5 = _mm_sub_pi32(mm5, *(v2si*)_pi32_2);
  mm4 = _mm_andnot_si64(mm4, *(v2si*)_pi32_4);
  mm5 = _mm_andnot_si64(mm5, *(v2si*)_pi32_4);
  mm4 = _mm_slli_pi32(mm4, 29);
  mm5 = _mm_slli_pi32(mm5, 29);
  v4sf sign_bit_cos;
  COPY_MM_TO_XMM(mm4, mm5, sign_bit_cos);
  sign_bit_sin = _mm_xor_ps(sign_bit_sin, swap_sign_bit_sin);
  v4sf z = _mm_mul_ps(x,x);
  y = *(v4sf*)_ps_coscof_p0;
  y = _mm_mul_ps(y, z);
  y = _mm_add_ps(y, *(v4sf*)_ps_coscof_p1);
  y = _mm_mul_ps(y, z);
  y = _mm_add_ps(y, *(v4sf*)_ps_coscof_p2);
  y = _mm_mul_ps(y, z);
  y = _mm_mul_ps(y, z);
  v4sf tmp = _mm_mul_ps(z, *(v4sf*)_ps_0p5);
  y = _mm_sub_ps(y, tmp);
  y = _mm_add_ps(y, *(v4sf*)_ps_1);
  v4sf y2 = *(v4sf*)_ps_sincof_p0;
  y2 = _mm_mul_ps(y2, z);
  y2 = _mm_add_ps(y2, *(v4sf*)_ps_sincof_p1);
  y2 = _mm_mul_ps(y2, z);
  y2 = _mm_add_ps(y2, *(v4sf*)_ps_sincof_p2);
  y2 = _mm_mul_ps(y2, z);
  y2 = _mm_mul_ps(y2, x);
  y2 = _mm_add_ps(y2, x);
  xmm3 = poly_mask;
  v4sf ysin2 = _mm_and_ps(xmm3, y2);
  v4sf ysin1 = _mm_andnot_ps(xmm3, y);
  y2 = _mm_sub_ps(y2,ysin2);
  y = _mm_sub_ps(y, ysin1);
  xmm1 = _mm_add_ps(ysin1,ysin2);
  xmm2 = _mm_add_ps(y,y2);
  *s = _mm_xor_ps(xmm1, sign_bit_sin);
  *c = _mm_xor_ps(xmm2, sign_bit_cos);
  _mm_empty();
}
#endif
}
