#include "astropulse.h"
#include "dm_chunk_parallel.h"
#include "ap_client_main.h"

/*
Iterate over Small DMs: (mainloop level 3)
	a) for each dm sign: (mainloop level 4)
		- Dechirp
		- For each Coadd Level: (mainloop level 5)
			- For each Power Bin (mainloop level 6)
				- Check for Pulse above threshold
				- Check/record Best Pulse
			- Coadd in place (reduces power array by half)
	b) reset sign to positive DM
*/
#ifdef PARALLEL
namespace Astropulse {
void Science::dm_chunk_proc()
{
	state_t& state = client.ap_shmem->ap_gdata.state;
	float*   power = &(client.power[0]);

	fprintf(stderr,"dm_chunk_proc() called\n");

        for (; state.dm_now<state.dm_chunk_small; state.dm_now++) {
//	  client.checkpoint();
          /* mainloop level 4
           * This loop iterates over the two dm signs, positive and negative
           */
          for (; state.dm_sign>=-1; state.dm_sign-=2) {
//            client.checkpoint();
            /* Dechirp the frequency domain data via multiplication by a dechirping
             * function.  Then perform the inverse fourier transform
             */
			dechirp(state.dm_chunk_small_now + state.dm_now, power, state.dm_sign);
            /* Graphics stuff */
  
            /* measuring the time that has passed since the program started running */
 //           updatecputime();
            /* mainloop level 5 * This loop iterates over coadd levels.
             */
	    for (int l=0; l<state.max_coadd; l++) {
              /* Uses ell, along with some members of client.state
               * Stores the result in state.frac_done.
               * The sole purpose of this computation is to print the fraction
               * done on the screen.
               */
              state.compute_fraction_done(l);
              /* boinc API boinc_fraction_done tells BOINC how much
               * of the current workunit has been completed.  This is
               * used to inform the core client GUI of the % complete.
               */
              boinc_fraction_done(state.frac_done);
   //			  boinc_ops_cumulative(state.frac_done*TOTAL_FLOPS,0);
			  boinc_ops_cumulative(state.frac_done*fpopsMult, 0);

              /* This is the size of the power array.
               * It starts at state.fft_len / 2, for ell = 0.  As ell gets larger, 
               * we perform coadds in place, and this reduces the effective 
               * size of the power array.
               * Note that from the very beginning, we ignore the second
               * half of the power bins.  Due to the overlap in ffts,
               * this second half is not necessary.
               */
              int dechirped_range_data_length = state.fft_len >> (l+1);
              /* mainloop level 6
              * This loop iterates over all of the power bins, searching for
              * a bin that exceeds the threshhold power.
              */
//              last_hit = -10;
	      double cur_cpu_time;
	      boinc_wu_cpu_time(cur_cpu_time);
	      client.ap_shmem->ap_gdata.cpu_time = cur_cpu_time;

              for (int m=0; m<dechirped_range_data_length; m++) {
  
		         if (power[m] > state.thresh[l]) {
		/* for debugging: check that the client randomizes these bits
		   if(state.data_chunk_now == 1359872 - 4096 && l == 3) {
		      for(int i = 16470; i < 16500; i++)
		        printf("data[%d][0] = %f, data[%d][1] = %f\n",
		           i, data[i][0], i, data[i][1]);
		   }
                 */
 
                /* If power exceeds threshhold, record a pulse.
                 * This involves first creating an ap_signal, which is defined in
                 * astropulse/server/db/ap_schema.h.  Then, we write the signal
                 * to pulse.out using Outfile::output_result.
                 */
                  if(pulse_limit_single && state.result_count_single >= pulse_limit_single)  // We've already found enough single pulses.  (Must be that we haven't found enough repeating pulses.)  Don't record this pulse.
                      continue;
                  ap_signal s;
                  time_series_b_t time_series;
                  time_series.build(power, m, dechirped_range_data_length, pow_int(2, l));
                  state.pulse_jd = state.jd0 + ((float)state.data_chunk_now)/((float)state.datasize)*(state.jd1-state.jd0);
  
                  s.dm= state.dm_sign*(state.dm_chunk_large_now+state.dm_chunk_small_now + state.dm_now); //dm
                  s.scale=l;                   // scale
                  s.peak_power=power[m];       // peak_power
                  s.fft_num=state.data_chunk_now*4;
                  s.peak_bin=state.data_chunk_now * 4 + (m << l); // index
                  s.time=state.pulse_jd;       // time
                  s.period=0.0;                // period
		  s.ffa_scale=0;		// ffa bin size
		  s.num_std_devs=-100.0;		// single pulse has no num_std_devs
                  // Interpolate betw. ra0 & ra1 using state.pulse_jd
                  s.ra = interpolate_ra(state.jd0, state.jd1, state.pulse_jd, state.ra0, state.ra1);
                  // Interpolate betw. dec0 and dec1
                  s.decl = interpolate_dec(state.jd0, state.jd1, state.pulse_jd, state.dec0, state.dec1);
                  s.time_series_len=time_series.length;
                  s.time_series=time_series.data; // WARNING why can we use an equal sign here?
                  s.time_series.encoding=_x_hex;
  		  if(add_signal(client.signal_vector, s)) state.result_count_single++;
                  // client.outfile.output_result(s);
                
                  if(pulse_limit_single && state.result_count_single >= pulse_limit_single && pulse_limit_rep && state.result_count_rep >= pulse_limit_rep) {
                    // Quit early
                    fprintf(stderr, "Found %d single pulses and %d repeating pulses, exiting.\n", state.result_count_single, state.result_count_rep);
		    printf("Found %d single pulses and %d repeating pulses, exiting.\n", state.result_count_single, state.result_count_rep);
	  	    fflush(stdout);
		    client.outfile.write_main(client.signal_vector);
                    client.outfile.finish_results();
                    client.finish();
                    main_timer.stop();
                    boinc_finish(0);
		    exit(0);
                  }
                }
  
                /* We record the best peak at each coadd level.  That is, there is
                 * one "best" value for each coadd level, for the entire run of
                 * the client.
                 */
                if (power[m] > state.best[l].peak_power) {
                  time_series_b_t time_series;
                  time_series.build(power, m, dechirped_range_data_length, pow_int(2, l));
                  state.pulse_jd = state.jd0 + (state.data_chunk_now)/(state.datasize)*(state.jd1-state.jd0);
                  state.best[l].fft_num=state.data_chunk_now*4;
                  state.best[l].peak_bin = state.data_chunk_now * 4 + (m << l);
                  state.best[l].peak_power = power[m];
                  state.best[l].scale = l;
                  state.best[l].dm = state.dm_sign*(state.dm_chunk_large_now+state.dm_chunk_small_now+state.dm_now);
                  state.best[l].period = 0.0;
		  state.best[l].ffa_scale = 0;
		  state.best[l].num_std_devs = -100.0;
		  state.best[l].ra = interpolate_ra(state.jd0, state.jd1, state.pulse_jd, state.ra0, state.ra1);
                  state.best[l].decl = interpolate_dec(state.jd0, state.jd1, state.pulse_jd, state.dec0, state.dec1);
                  state.best[l].time = state.pulse_jd;
		  state.best[l].time_series_len = time_series.length;
		  assert(time_series.length <= MAXLENGTH_OF_TS);  // Don't want an overflow
		  // This memcpy works because state_signal.time_series is an array, and so is
		  // time_series.data.  (ap_signal.time_series would be an sqlblob.)
		  memcpy(state.best[l].time_series, &(time_series.data[0]), time_series.length);
                }
              } /* End mainloop level 6, bin number (m) */

#if !NDEBUG
							debug_loop(5, 10, debug_loop_msg);
#endif
#if USE_GPU
	}//R:end of single pulse search
#endif
	      /* When l is at two specific values (called state.fold_level_large and
               * state.fold_level small), we are ready to contribute data to the fast
               * folding algorithm.  
               * 
               * Large scale version: This happens only once per (dm_chunk_large, data_chunk) pair.
               * So we only hit one out of every 2^(state.fold_level_large)
               * dms in this way (henceforth 2^(s.fll)).  This is why we do it at a fold_level
               * such that 2^(s.fll) is equal to state.dm_chunk_large: it's not meaningful to look
               * at data at a better time resolution than that provided by the dm resolution.
               *
               * We also look at a finer resolution (below), namely 2^(state.fold_level_small).  
               * In this case, we contribute data more often, and there is more of it.  To compensate for this,
               * we must cut off the power array at some number of bytes, state.fold_buf_bytes_short.
               */
              // Temporary fix
              if ((l==state.fold_level_large) && (state.dm_now == fold_dm) && (state.dm_chunk_small_now == fold_dm_chunk_small)) {
                /* Save for folding */
 		// positive dms
			if(state.dm_sign == 1) { 
					memcpy(&client.fold_buf_long_pos[state.fold_buf_loc_long_pos], power,
							   sizeof(float)*(state.fft_len>>(l+1)));
					state.fold_buf_loc_long_pos += (state.fft_len>>(l+1));
							  // On the last data chunk, do one extra memcpy.
				  // This is because the number of iterations in the data chunk loop is one less than enough
				  // to make the long fold buffer fill up, due to the overlap between data chunks.
				  // Note that datasize is measured in bytes, and increments by fft_len/8.
				  if(state.data_chunk_now == (state.datasize - state.fft_len/4)) {
							memcpy(&client.fold_buf_long_pos[state.fold_buf_loc_long_pos], power,
								 sizeof(float)*(state.fft_len>>(l+1)));
							state.fold_buf_loc_long_pos += (state.fft_len>>(l+1));
				  }
			// identical code, but for negative dms
			} else if(state.dm_sign == -1) { 
				  memcpy(&client.fold_buf_long_neg[state.fold_buf_loc_long_neg], power,
		  			sizeof(float)*(state.fft_len>>(l+1)));
						  state.fold_buf_loc_long_neg += (state.fft_len>>(l+1));
				  if(state.data_chunk_now == (state.datasize - state.fft_len/4)) {
						memcpy(&client.fold_buf_long_neg[state.fold_buf_loc_long_neg], power,
      						sizeof(float)*(state.fft_len>>(l+1)));
						state.fold_buf_loc_long_neg += (state.fft_len>>(l+1));
				  }
			} else { 
							fprintf(stderr, "Error in ap_client_main.cpp: bad dm_sign value");
							exit(-1);
			} //dm_sign
	   } //if fold level large
#if !NDEBUG
	      debug_loop(5, 11, debug_loop_msg);
#endif
	      /* More frequently, we copy the power array into the short fold buffer.
	       * Note that the condition  state.data_chunk_now < state.fold_buf_bytes_short,
	       * even though each data chunk extends beyond the start of the next one.
 	       */
   	      if ((l==state.fold_level_small) && (state.dm_now == fold_dm) 
					&& (state.data_chunk_now < state.fold_buf_bytes_short)) 
		  {
  	            /* Save for folding in short time series, for finding short period repeating pulses. */
				if(state.dm_sign == 1) {
						  memcpy(&client.fold_buf_short_pos[state.fold_buf_loc_short_pos], power,
										sizeof(float)*(state.fft_len>>(l+1)));
						  state.fold_buf_loc_short_pos += (state.fft_len>>(l+1));
				} else if(state.dm_sign == -1) {
						  memcpy(&client.fold_buf_short_neg[state.fold_buf_loc_short_neg], power,
										sizeof(float)*(state.fft_len>>(l+1)));
						  state.fold_buf_loc_short_neg += (state.fft_len>>(l+1));
				}
		  }
#if !NDEBUG
	      debug_loop(5, 12, debug_loop_msg);
#endif
              /* coadd_in_place is defined in ap_science.cpp
               * Its purpose is to halve the size of the power array by adding
               * array elements in consecutive pairs.
               * The second argument is the initial array size.  Note that
               * 
               */
#if USE_GPU
		  GPU_coadd_in_place(gpupower, state.fft_len>>(l+1));
#else
		  coadd_in_place(power, state.fft_len>>(l+1));
#endif
#if !NDEBUG
	      debug_loop(5, 13, debug_loop_msg);
#endif
            } /* End mainloop level 5, coadd level (l) */
#if !NDEBUG
	    debug_loop(4, 10, debug_loop_msg);


            // client.checkpoint(); Can only checkpoint at beginning of for loop
	    debug_loop(4, 11, debug_loop_msg);
#endif
          } /* End mainloop level 4, dm sign (s) */
#if !NDEBUG
	  debug_loop(3, 10, debug_loop_msg);
#endif
          state.dm_sign = 1;
#if !NDEBUG
	  debug_loop(3, 11, debug_loop_msg);
#endif
        } /* End mainloop level 3, inner dm (k) */
	return;
}
} //namespace Astropulse
#endif  // PARALELL